@@ -82,14 +82,13 @@ struct domain_arg {
 struct queued_task_ctx {
 	s32 pid;
 	s32 cpu; /* CPU where the task is running */
-	u64 flags; /* task enqueue flags */
+	u64 nr_cpus_allowed; /* Number of CPUs that the task can use */
+	u64 flags; /* Task enqueue flags */
+	u64 start_ts; /* Timestamp since last time the task ran on a CPU */
+	u64 stop_ts; /* Timestamp since last time the task released a CPU */
 	u64 exec_runtime; /* Total cpu time since last sleep */
-	u64 sum_exec_runtime; /* Total cpu time */
-	u64 nvcsw; /* Total amount of voluntary context switches */
 	u64 weight; /* Task static priority */
-	u64 slice; /* Time slice budget */
 	u64 vtime; /* Current task's vruntime */
-	u64 cpumask_cnt; /* cpumask generation counter */
 };
 
 /*
@@ -104,7 +103,6 @@ struct dispatched_task_ctx {
 	u64 flags; /* task enqueue flags */
 	u64 slice_ns; /* time slice assigned to the task (0=default) */
 	u64 vtime; /* task deadline / vruntime */
-	u64 cpumask_cnt; /* cpumask generation counter */
 };
 
 #endif /* __INTF_H */
