@@ -110,8 +110,8 @@ impl Cpumask {
             }
             tmp_str
         };
-        let byte_vec = hex::decode(&hex_str)
-            .with_context(|| format!("Failed to parse cpumask: {}", cpumask))?;
+        let byte_vec =
+            hex::decode(&hex_str).with_context(|| format!("Failed to parse cpumask: {cpumask}"))?;
 
         let mut mask = bitvec![u64, Lsb0; 0; *NR_CPU_IDS];
         for (index, &val) in byte_vec.iter().rev().enumerate() {
@@ -279,7 +279,9 @@ impl Cpumask {
     /// Write out a CPU mask to a raw memory pointer. We normally use this as part of updating
     /// the CPU masks on the BPF side.
     ///
-    /// SAFETY: The caller must enforce that the bpfptr is valid and points to a valid region of
+    /// # Safety
+    ///
+    /// The caller must enforce that the bpfptr is valid and points to a valid region of
     /// writable BPF memory that represents a CPU mask. This function is not thread safe.
     pub unsafe fn write_to_ptr(&self, bpfptr: *mut u64, len: usize) -> Result<()> {
         let cpumask_slice = self.as_raw_slice();
@@ -323,8 +325,8 @@ impl Cpumask {
         // The rest in descending order.
         for submask in masks.iter().rev() {
             match case {
-                'x' => write!(f, ",{:08x}", submask)?,
-                'X' => write!(f, ",{:08X}", submask)?,
+                'x' => write!(f, ",{submask:08x}")?,
+                'X' => write!(f, ",{submask:08X}")?,
                 _ => unreachable!(),
             }
         }
