@@ -51,10 +51,7 @@ pub struct EnergyModel {
 
 impl EnergyModel {
     pub fn has_energy_model() -> bool {
-        match get_pd_paths() {
-            Ok(_) => true,
-            Err(_) => false,
-        }
+        get_pd_paths().is_ok()
     }
 
     /// Build a complete EnergyModel
@@ -78,7 +75,7 @@ impl EnergyModel {
     pub fn get_pd_by_cpu_id(&self, cpu_id: usize) -> Option<&PerfDomain> {
         for (_, pd) in self.perf_doms.iter() {
             if pd.span.test_cpu(cpu_id) {
-                return Some(&pd);
+                return Some(pd);
             }
         }
         None
@@ -120,7 +117,7 @@ impl PerfDomain {
         let util = clamp(util, 0.0, 100.0);
         let (perf_max, _) = self.perf_table.last_key_value().unwrap();
         let perf_max = *perf_max as f32;
-        let req_perf = (perf_max as f32 * (util / 100.0)) as usize;
+        let req_perf = (perf_max * (util / 100.0)) as usize;
         for (perf, ps) in self.perf_table.iter() {
             if *perf >= req_perf {
                 return Some(ps);
@@ -172,7 +169,7 @@ impl PartialEq for PerfState {
 impl fmt::Display for EnergyModel {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         for (_, pd) in self.perf_doms.iter() {
-            writeln!(f, "{:#}", pd)?;
+            writeln!(f, "{pd:#}")?;
         }
         Ok(())
     }
@@ -183,7 +180,7 @@ impl fmt::Display for PerfDomain {
         writeln!(f, "# perf domain: {:#}, cpus: {:#}", self.id, self.span)?;
         writeln!(f, "cost, frequency, inefficient, performance, power")?;
         for (_, ps) in self.perf_table.iter() {
-            writeln!(f, "{:#}", ps)?;
+            writeln!(f, "{ps:#}")?;
         }
         Ok(())
     }
@@ -238,7 +235,7 @@ fn get_pd_paths() -> Result<Vec<(usize, String)>> {
 }
 
 fn get_em_root() -> Result<String> {
-    if *ROOT_PREFIX == "" {
+    if ROOT_PREFIX.is_empty() {
         let root = compat::debugfs_mount().unwrap().join("energy_model");
         Ok(root.display().to_string())
     } else {
