@@ -17,8 +17,9 @@ struct scx_atq {
 
 typedef __arena scx_atq, scx_atq_t;
 
-u64 scx_atq_create_internal(bool fifo);
-#define scx_atq_create(fifo) (scx_atq_create_internal((fifo)))
+u64 scx_atq_create_internal(bool fifo, size_t capacity);
+#define scx_atq_create(fifo) scx_atq_create_internal((fifo), SCX_ATQ_MAX_CAPACITY)
+#define scx_atq_create_size(fifo, capacity) scx_atq_create_internal((fifo), (capacity))
 int scx_atq_insert(scx_atq *atq_ptr, u64 taskc_ptr);
 int scx_atq_insert_vtime(scx_atq_t *atq, u64 taskc_ptr, u64 vtime);
 int scx_atq_nr_queued(scx_atq_t *atq);
