@@ -13,6 +13,7 @@ use scx_p2dq::SchedulerOpts as P2dqOpts;
 use scx_userspace_arena::alloc::Allocator;
 use scx_userspace_arena::alloc::HeapAllocator;
 use scx_utils::build_id;
+use scx_utils::compat;
 use scx_utils::compat::tracefs_mount;
 use scx_utils::init_libbpf_logging;
 use scx_utils::scx_ops_attach;
@@ -140,6 +141,8 @@ pub enum RequiresPpid {
 pub struct KprobeRandomDelays {
     pub kprobes: Vec<String>,
     pub freq: f64,
+    pub min_us: u64,
+    pub max_us: u64,
 }
 
 #[derive(Debug)]
@@ -196,6 +199,8 @@ impl Scheduler {
                 [bpf_intf::chaos_stat_idx_CHAOS_STAT_TRAIT_DEGRADATION as usize],
             chaos_excluded: stats[bpf_intf::chaos_stat_idx_CHAOS_STAT_CHAOS_EXCLUDED as usize],
             chaos_skipped: stats[bpf_intf::chaos_stat_idx_CHAOS_STAT_CHAOS_SKIPPED as usize],
+            kprobe_random_delays: stats
+                [bpf_intf::chaos_stat_idx_CHAOS_STAT_KPROBE_RANDOM_DELAYS as usize],
             timer_kicks: stats[bpf_intf::chaos_stat_idx_CHAOS_STAT_TIMER_KICKS as usize],
         }
     }
@@ -381,7 +386,7 @@ impl Builder<'_> {
                 skel.progs
                     .generic
                     .attach_kprobe(false, k)
-                    .context(format!("Failed to attach kprobe {:?}", k))?,
+                    .context(format!("Failed to attach kprobe {k:?}"))?,
             );
         }
         Ok(links)
@@ -415,32 +420,41 @@ impl Builder<'_> {
         let mut open_skel = scx_ops_open!(skel_builder, open_object, chaos)?;
         scx_p2dq::init_open_skel!(&mut open_skel, self.p2dq_opts, self.verbose)?;
 
+        let rodata = open_skel.maps.rodata_data.as_mut().unwrap();
+
         // TODO: figure out how to abstract waking a CPU in enqueue properly, but for now disable
         // this codepath
-        open_skel.maps.rodata_data.select_idle_in_enqueue = false;
+        rodata.p2dq_config.select_idle_in_enqueue = MaybeUninit::new(false);
+
+        if self.p2dq_opts.queued_wakeup {
+            open_skel.struct_ops.chaos_mut().flags |= *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP;
+        }
+        open_skel.struct_ops.chaos_mut().flags |= *compat::SCX_OPS_KEEP_BUILTIN_IDLE;
 
         match self.requires_ppid {
             None => {
-                open_skel.maps.rodata_data.ppid_targeting_ppid = -1;
+                rodata.ppid_targeting_ppid = -1;
             }
             Some(RequiresPpid::ExcludeParent(p)) => {
-                open_skel.maps.rodata_data.ppid_targeting_inclusive = false;
-                open_skel.maps.rodata_data.ppid_targeting_ppid = p.as_raw();
+                rodata.ppid_targeting_inclusive = false;
+                rodata.ppid_targeting_ppid = p.as_raw();
             }
             Some(RequiresPpid::IncludeParent(p)) => {
-                open_skel.maps.rodata_data.ppid_targeting_inclusive = true;
-                open_skel.maps.rodata_data.ppid_targeting_ppid = p.as_raw();
+                rodata.ppid_targeting_inclusive = true;
+                rodata.ppid_targeting_ppid = p.as_raw();
             }
         };
 
         if let Some(kprobe_random_delays) = &self.kprobe_random_delays {
-            open_skel.maps.rodata_data.kprobe_delays_freq_frac32 =
+            rodata.kprobe_delays_freq_frac32 =
                 (kprobe_random_delays.freq * 2_f64.powf(32_f64)) as u32;
+            rodata.kprobe_delays_min_ns = kprobe_random_delays.min_us * 1000;
+            rodata.kprobe_delays_max_ns = kprobe_random_delays.max_us * 1000;
         }
 
         // Set up the frequency array. The first element means nothing, so should be what's
         // required to add up to 100%. The rest should be cumulative frequencies.
-        let freq_array = &mut open_skel.maps.rodata_data.trait_delay_freq_frac32;
+        let freq_array = &mut rodata.trait_delay_freq_frac32;
         freq_array.fill(0);
         for tr in &self.traits {
             let kind = tr.kind();
@@ -467,7 +481,7 @@ impl Builder<'_> {
 
         debug!(
             "frequencies calculated as: {:?}",
-            open_skel.maps.rodata_data.trait_delay_freq_frac32
+            rodata.trait_delay_freq_frac32
         );
 
         for tr in &self.traits {
@@ -477,26 +491,25 @@ impl Builder<'_> {
                     min_us,
                     max_us,
                 } => {
-                    open_skel.maps.rodata_data.random_delays_min_ns = min_us * 1000;
-                    open_skel.maps.rodata_data.random_delays_max_ns = max_us * 1000;
+                    rodata.random_delays_min_ns = min_us * 1000;
+                    rodata.random_delays_max_ns = max_us * 1000;
                 }
                 Trait::CpuFreq {
                     frequency: _,
                     min_freq,
                     max_freq,
                 } => {
-                    open_skel.maps.rodata_data.cpu_freq_min = *min_freq;
-                    open_skel.maps.rodata_data.cpu_freq_max = *max_freq;
+                    rodata.cpu_freq_min = *min_freq;
+                    rodata.cpu_freq_max = *max_freq;
                     // Don't let p2dq control frequency
-                    open_skel.maps.rodata_data.freq_control = false;
+                    rodata.p2dq_config.freq_control = MaybeUninit::new(false);
                 }
                 Trait::PerfDegradation {
                     frequency,
                     degradation_frac7,
                 } => {
-                    open_skel.maps.rodata_data.degradation_freq_frac32 =
-                        (frequency * 2_f64.powf(32_f64)) as u32;
-                    open_skel.maps.rodata_data.degradation_frac7 = *degradation_frac7;
+                    rodata.degradation_freq_frac32 = (frequency * 2_f64.powf(32_f64)) as u32;
+                    rodata.degradation_frac7 = *degradation_frac7;
                 }
             }
         }
@@ -602,12 +615,20 @@ pub struct PerfDegradationArgs {
 #[derive(Debug, Parser)]
 pub struct KprobeArgs {
     /// Introduce random delays in the scheduler whenever a provided kprobe is hit.
-    #[clap(long, num_args = 1.., value_parser)]
+    #[clap(long, num_args = 1.., value_parser, requires = "kprobe_random_delay_min_us")]
     pub kprobes_for_random_delays: Vec<String>,
 
-    /// Chance of kprobe random delays. Must be between 0 and 1.
+    /// Chance of kprobe random delays. Must be between 0 and 1. [default=0.1]
     #[clap(long, requires = "kprobes_for_random_delays")]
     pub kprobe_random_delay_frequency: Option<f64>,
+
+    /// Minimum time to add for kprobe random delay.
+    #[clap(long, requires = "kprobe_random_delay_max_us")]
+    pub kprobe_random_delay_min_us: Option<u64>,
+
+    /// Maximum time to add for kprobe random delay.
+    #[clap(long, requires = "kprobes_for_random_delays")]
+    pub kprobe_random_delay_max_us: Option<u64>,
 }
 
 /// scx_chaos: A general purpose sched_ext scheduler designed to amplify race conditions
@@ -757,9 +778,13 @@ impl<'a> Iterator for BuilderIterator<'a> {
                 KprobeArgs {
                     kprobes_for_random_delays,
                     kprobe_random_delay_frequency,
+                    kprobe_random_delay_min_us: Some(min_us),
+                    kprobe_random_delay_max_us: Some(max_us),
                 } if !kprobes_for_random_delays.is_empty() => Some(KprobeRandomDelays {
                     kprobes: kprobes_for_random_delays.clone(),
                     freq: kprobe_random_delay_frequency.unwrap_or(0.1),
+                    min_us: *min_us,
+                    max_us: *max_us,
                 }),
                 _ => None,
             };
@@ -858,7 +883,7 @@ pub fn run(args: Args) -> Result<()> {
     });
 
     if let Some(pid) = args.pid {
-        info!("Monitoring process with PID: {}", pid);
+        info!("Monitoring process with PID: {pid}");
 
         let is_process_running = |pid: libc::pid_t| -> bool {
             unsafe {
