@@ -72,7 +72,7 @@ struct pick_ctx {
 static __always_inline
 bool init_idle_i_mask(struct pick_ctx *ctx, const struct cpumask *idle_cpumask)
 {
-	if (!ctx->taskc->is_affinitized)
+	if (!test_task_flag(ctx->taskc, LAVD_FLAG_IS_AFFINITIZED))
 		ctx->i_mask = idle_cpumask;
 	else {
 		struct bpf_cpumask *_i_mask = ctx->cpuc_cur->tmp_i_mask;
@@ -102,7 +102,7 @@ bool init_ao_masks(struct pick_ctx *ctx)
 	if (!ctx->cpuc_cur)
 		return false;
 
-	if (!ctx->taskc->is_affinitized) {
+	if (!test_task_flag(ctx->taskc, LAVD_FLAG_IS_AFFINITIZED)) {
 		ctx->a_mask = ctx->active;
 		ctx->o_mask = ctx->ovrflw;
 		ctx->a_empty = ctx->o_empty = false;
@@ -256,6 +256,27 @@ s32 cpumask_any_dsitribute(struct pick_ctx *ctx)
 	return -ENOENT;
 }
 
+static
+s32 pick_random_cpu(struct pick_ctx *ctx)
+{
+	/*
+	 * Pick a less loaded CPU using the random of two choices technique.
+	 */
+	s32 cpu0 = cpumask_any_dsitribute(ctx);
+	s32 cpu1 = cpumask_any_dsitribute(ctx);
+	struct cpu_ctx *cpuc0, *cpuc1;
+
+	if (cpu0 == cpu1 && cpu0 != -ENOENT)
+		return cpu0;
+
+	cpuc0 = get_cpu_ctx_id(cpu0);
+	cpuc1 = get_cpu_ctx_id(cpu1);
+	if (!cpuc0 || !cpuc1)
+		return ctx->prev_cpu;
+
+	return (cpuc0->cur_sc_util < cpuc1->cur_sc_util) ? cpu0 : cpu1;
+}
+
 static
 s32 find_sticky_cpu_at_cpdom(struct pick_ctx *ctx, s32 sticky_cpu, s64 sticky_cpdom)
 {
@@ -299,7 +320,7 @@ bool can_run_on_cpu(struct pick_ctx *ctx, s32 cpu)
 	struct bpf_cpumask *a_mask;
 	struct bpf_cpumask *o_mask;
 
-	if (!ctx->taskc->is_affinitized)
+	if (!test_task_flag(ctx->taskc, LAVD_FLAG_IS_AFFINITIZED))
 		return true;
 
 	if (!bpf_cpumask_test_cpu(cpu, ctx->p->cpus_ptr))
@@ -320,7 +341,7 @@ bool can_run_on_domain(struct pick_ctx *ctx, s64 cpdom)
 	struct cpdom_ctx *cpdc;
 	struct bpf_cpumask *cpd_mask, *a_mask, *o_mask;
 
-	if (!ctx->taskc->is_affinitized)
+	if (!test_task_flag(ctx->taskc, LAVD_FLAG_IS_AFFINITIZED))
 		return true;
 
 	cpd_mask = MEMBER_VPTR(cpdom_cpumask, [cpdom]);
@@ -629,7 +650,7 @@ s32 pick_idle_cpu(struct pick_ctx *ctx, bool *is_idle)
 	 * overflow set.
 	 */
 	if (sticky_cpdom < 0) {
-		cpu = cpumask_any_dsitribute(ctx);
+		cpu = pick_random_cpu(ctx);
 		goto unlock_out;
 	}
 	/* NOTE: There is a sticky domain. */
@@ -802,12 +823,6 @@ skip_fully_idle_neighbor:
 err_out:
 	cpu = -ENOENT;
 unlock_out:
-	if (idle_smtmask)
-		scx_bpf_put_idle_cpumask(idle_smtmask);
-	if (idle_cpumask)
-		scx_bpf_put_idle_cpumask(idle_cpumask);
-	bpf_rcu_read_unlock();
-
 	if (sticky_cpdom < 0) {
 		struct cpu_ctx *cpuc;
 		cpuc = get_cpu_ctx_id(cpu >= 0 ? cpu : ctx->prev_cpu);
@@ -815,17 +830,29 @@ unlock_out:
 			ctx->cpdom_id = cpuc->cpdom_id;
 	} else
 		ctx->cpdom_id = sticky_cpdom;
+
+	if (cpu == -ENOENT)
+		cpu = pick_random_cpu(ctx);
+
+	if (idle_smtmask)
+		scx_bpf_put_idle_cpumask(idle_smtmask);
+	if (idle_cpumask)
+		scx_bpf_put_idle_cpumask(idle_cpumask);
+	bpf_rcu_read_unlock();
+
 	return cpu;
 }
 
 static
 s64 pick_proper_dsq(const struct task_struct *p, struct task_ctx *taskc,
-		    s32 task_cpu, s32 *cpu, bool *is_idle)
+		    s32 task_cpu, s32 *cpu, bool *is_idle,
+		    struct cpu_ctx *cpuc_cur)
 {
 	struct pick_ctx ictx = {
 		.p = p,
 		.taskc = taskc,
 		.prev_cpu = task_cpu,
+		.cpuc_cur = cpuc_cur,
 		.wake_flags = 0,
 		.cpdom_id = -ENOMEM,
 	};
