@@ -8,6 +8,11 @@
  * To be included to the main.bpf.c
  */
 
+/*
+ * Flag to represent whether the scheduler is being monitored or not.
+ */
+volatile bool is_monitored;
+
 /*
  * Introspection commands
  */
@@ -22,12 +27,17 @@ static __always_inline
 int submit_task_ctx(struct task_struct *p, struct task_ctx *taskc, u32 cpu_id)
 {
 	struct cpu_ctx *cpuc;
+	struct cpdom_ctx *cpdomc;
 	struct msg_task_ctx *m;
 
 	cpuc = get_cpu_ctx_id(cpu_id);
 	if (!cpuc)
 		return -EINVAL;
 
+	cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpuc->cpdom_id]);
+	if (!cpdomc)
+		return -EINVAL;
+
 	m = bpf_ringbuf_reserve(&introspec_msg, sizeof(*m), 0);
 	if (!m)
 		return -ENOMEM;
@@ -38,16 +48,19 @@ int submit_task_ctx(struct task_struct *p, struct task_ctx *taskc, u32 cpu_id)
 	m->taskc_x.static_prio = get_nice_prio(p);
 	m->taskc_x.cpu_util = s2p(cpuc->avg_util);
 	m->taskc_x.cpu_sutil = s2p(cpuc->avg_sc_util);
-	m->taskc_x.cpu_id = cpu_id;
+	m->taskc_x.rerunnable_interval = time_delta(taskc->last_quiescent_clk, taskc->last_runnable_clk);
 	m->taskc_x.avg_lat_cri = sys_stat.avg_lat_cri;
 	m->taskc_x.thr_perf_cri = sys_stat.thr_perf_cri;
 	m->taskc_x.nr_active = sys_stat.nr_active;
 	m->taskc_x.cpuperf_cur = cpuc->cpuperf_cur;
+	/* Refactor this when per-cpu DSQs are added */
+	m->taskc_x.dsq_id = cpdomc->id;
+	m->taskc_x.dsq_consume_lat = cpdomc->dsq_consume_lat;
 
 	m->taskc_x.stat[0] = is_lat_cri(taskc) ? 'L' : 'R';
 	m->taskc_x.stat[1] = is_perf_cri(taskc) ? 'H' : 'I';
 	m->taskc_x.stat[2] = cpuc->big_core ? 'B' : 'T';
-	m->taskc_x.stat[3] = is_greedy(taskc) ? 'G' : 'E';
+	m->taskc_x.stat[3] = test_task_flag(taskc, LAVD_FLAG_IS_GREEDY)? 'G' : 'E';
 	m->taskc_x.stat[4] = '\0';
 
 	__builtin_memcpy_inline(&m->taskc, taskc, sizeof(m->taskc));
@@ -95,6 +108,9 @@ static void proc_introspec_sched_n(struct task_struct *p,
 static void try_proc_introspec_cmd(struct task_struct *p,
 				   struct task_ctx *taskc)
 {
+	if (!is_monitored)
+		return;
+
 	switch(intrspc.cmd) {
 	case LAVD_CMD_SCHED_N:
 		proc_introspec_sched_n(p, taskc);
