@@ -221,33 +221,6 @@ const volatile u64	slice_max_ns = LAVD_SLICE_MAX_NS_DFL;
 #include "sys_stat.bpf.c"
 #include "lat_cri.bpf.c"
 
-static void reset_suspended_duration(struct cpu_ctx *cpuc)
-{
-	if (cpuc->online_clk > cpuc->offline_clk)
-		cpuc->offline_clk = cpuc->online_clk;
-}
-
-static u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
-{
-	/*
-	 * When a system is suspended, a task is also suspended in a running
-	 * stat on the CPU. Hence, we subtract the suspended duration when it
-	 * resumes.
-	 */
-
-	u64 duration = 0;
-
-	if (cpuc->online_clk > cpuc->offline_clk) {
-		duration = time_delta(cpuc->online_clk, cpuc->offline_clk);
-		/*
-		 * Once calculated, reset the duration to zero.
-		 */
-		cpuc->offline_clk = cpuc->online_clk;
-	}
-
-	return duration;
-}
-
 static void advance_cur_logical_clk(struct task_struct *p)
 {
 	u64 vlc, clc, ret_clc;
@@ -283,13 +256,79 @@ static void advance_cur_logical_clk(struct task_struct *p)
 	}
 }
 
-static u64 calc_time_slice(struct task_ctx *taskc)
+static u64 calc_time_slice(struct task_ctx *taskc, struct cpu_ctx *cpuc)
 {
-	if (!taskc)
+	/*
+	 * Calculate the time slice of @taskc to run on @cpuc.
+	 */
+	if (!taskc || !cpuc)
 		return LAVD_SLICE_MAX_NS_DFL;
 
-	taskc->slice_ns = sys_stat.slice;
-	return taskc->slice_ns;
+	/*
+	 * If the task's avg_runtime is greater than the regular time slice
+	 * (i.e., taskc->avg_runtime > sys_stat.slice), that means the task
+	 * could be scheduled out due to a shorter time slice than required.
+	 * In this case, let's consider boosting task's time slice.
+	 *
+	 * However, if there are pinned tasks waiting to run on this CPU,
+	 * we do not boost the task's time slice to avoid delaying the pinned
+	 * task that cannot be run on another CPU.
+	 */
+	if (!no_slice_boost && !cpuc->nr_pinned_tasks &&
+	    (taskc->avg_runtime >= sys_stat.slice)) {
+		/*
+		 * When the system is not heavily loaded, so it can serve all
+		 * tasks within the targeted latency (slice_max_ns <=
+		 * sys_stat.slice), we fully boost task's time slice.
+		 *
+		 * Let's set the task's time slice to its avg_runtime
+		 * (+ some bonus) to reduce unnecessary involuntary context
+		 * switching.
+		 *
+		 * Even in this case, we want to limit the maximum time slice
+		 * to LAVD_SLICE_BOOST_MAX (not infinite) because we want to
+		 * revisit if the task is placed on the best CPU at least
+		 * every LAVD_SLICE_BOOST_MAX interval.
+		 */
+		if (can_boost_slice()) {
+			/*
+			 * Add a bit of bonus so that a task, which takes a
+			 * bit longer than average, can still finish the job.
+			 */
+			u64 s = taskc->avg_runtime + LAVD_SLICE_BOOST_BONUS;
+			taskc->slice = clamp(s, slice_min_ns,
+					     LAVD_SLICE_BOOST_MAX);
+			set_task_flag(taskc, LAVD_FLAG_SLICE_BOOST);
+			return taskc->slice;
+		}
+
+		/*
+		 * When the system is under high load, we will boost the time
+		 * slice of only latency-critical tasks, which are likely in
+		 * the middle of a task chain. Also, increase the time slice
+		 * proportionally to the latency criticality up to 2x the
+		 * regular time slice.
+		 */
+		if (taskc->lat_cri > sys_stat.avg_lat_cri) {
+			u64 b = (sys_stat.slice * taskc->lat_cri) /
+				(sys_stat.avg_lat_cri + 1);
+			u64 s = sys_stat.slice + b;
+			taskc->slice = clamp(s, slice_min_ns,
+					     min(taskc->avg_runtime,
+						 sys_stat.slice * 2));
+
+			set_task_flag(taskc, LAVD_FLAG_SLICE_BOOST);
+			return taskc->slice;
+		}
+	}
+
+	/*
+	 * If slice boost is either not possible, not necessary, or not
+	 * eligible, assign the regular time slice.
+	 */
+	taskc->slice = sys_stat.slice;
+	reset_task_flag(taskc, LAVD_FLAG_SLICE_BOOST);
+	return taskc->slice;
 }
 
 static void update_stat_for_running(struct task_struct *p,
@@ -297,25 +336,7 @@ static void update_stat_for_running(struct task_struct *p,
 				    struct cpu_ctx *cpuc, u64 now)
 {
 	u64 wait_period, interval;
-
-	/*
-	 * If the sched_ext core directly dispatched a task, calculating the
-	 * task's deadline and time slice was also skipped. In this case, we
-	 * set the deadline and time slice here.
-	 *
-	 * Note that this is necessary when the kernel does not support
-	 * SCX_OPS_ENQ_MIGRATION_DISABLED or SCX_OPS_ENQ_MIGRATION_DISABLED
-	 * is not turned on.
-	 */
-	if (p->scx.slice == SCX_SLICE_DFL) {
-		p->scx.dsq_vtime = READ_ONCE(cur_logical_clk);
-		p->scx.slice = calc_time_slice(taskc);
-	}
-
-	/*
-	 * Update the current logical clock.
-	 */
-	advance_cur_logical_clk(p);
+	struct cpu_ctx *prev_cpuc;
 
 	/*
 	 * Since this is the start of a new schedule for @p, we update run
@@ -324,14 +345,27 @@ static void update_stat_for_running(struct task_struct *p,
 	if (have_scheduled(taskc)) {
 		wait_period = time_delta(now, taskc->last_quiescent_clk);
 		interval = taskc->avg_runtime + wait_period;
-		taskc->run_freq = calc_avg_freq(taskc->run_freq, interval);
+		if (interval > 0)
+			taskc->run_freq = calc_avg_freq(taskc->run_freq, interval);
 	}
 
+	/*
+	 * Collect additional information when the scheduler is monitored.
+	 */
+	if (is_monitored) {
+		taskc->resched_interval = time_delta(now,
+						     taskc->last_running_clk);
+	}
+	taskc->prev_cpu_id = taskc->cpu_id;
+	taskc->cpu_id = cpuc->cpu_id;
+
 	/*
 	 * Update task state when starts running.
 	 */
-	taskc->wakeup_ft = 0;
+	reset_task_flag(taskc, LAVD_FLAG_IS_WAKEUP);
+	reset_task_flag(taskc, LAVD_FLAG_IS_SYNC_WAKEUP);
 	taskc->last_running_clk = now;
+	taskc->last_measured_clk = now;
 
 	/*
 	 * Reset task's lock and futex boost count
@@ -349,7 +383,7 @@ static void update_stat_for_running(struct task_struct *p,
 	cpuc->nr_sched++;
 
 	/*
-	 * Update per-CPU performanc criticality information
+	 * Update per-CPU performance criticality information
 	 * for every-scheduled tasks.
 	 */
 	if (have_little_core) {
@@ -360,6 +394,14 @@ static void update_stat_for_running(struct task_struct *p,
 		cpuc->sum_perf_cri += taskc->perf_cri;
 	}
 
+	/*
+	 * Update running task's information for preemption
+	 */
+	cpuc->flags = taskc->flags;
+	cpuc->lat_cri = taskc->lat_cri;
+	cpuc->running_clk = now;
+	cpuc->est_stopping_clk = get_est_stopping_clk(taskc, now);
+
 	/*
 	 * Update statistics information.
 	 */
@@ -369,10 +411,9 @@ static void update_stat_for_running(struct task_struct *p,
 	if (is_perf_cri(taskc))
 		cpuc->nr_perf_cri++;
 
-	if (taskc->dsq_id != cpuc->cpdom_id) {
-		taskc->dsq_id = cpuc->cpdom_id;
+	prev_cpuc = get_cpu_ctx_id(taskc->prev_cpu_id);
+	if (prev_cpuc && prev_cpuc->cpdom_id != cpuc->cpdom_id)
 		cpuc->nr_x_migration++;
-	}
 
 	/*
 	 * It is clear there is no need to consider the suspended duration
@@ -381,29 +422,53 @@ static void update_stat_for_running(struct task_struct *p,
 	reset_suspended_duration(cpuc);
 }
 
+static void account_task_runtime(struct task_struct *p,
+				 struct task_ctx *taskc,
+				 struct cpu_ctx *cpuc,
+				 u64 now)
+{
+	u64 sus_dur, runtime, svc_time, sc_time;
+
+	/*
+	 * Since task execution can span one or more sys_stat intervals,
+	 * we update task and CPU's statistics at every tick interval and
+	 * update_stat_for_stopping(). It is essential to account for
+	 * the load of long-running tasks properly. So, we add up only the
+	 * execution duration since the last measured time.
+	 */
+	sus_dur = get_suspended_duration_and_reset(cpuc);
+	runtime = time_delta(now, taskc->last_measured_clk + sus_dur);
+	svc_time = runtime / p->scx.weight;
+	sc_time = scale_cap_freq(runtime, cpuc->cpu_id);
+
+	WRITE_ONCE(cpuc->tot_svc_time, cpuc->tot_svc_time + svc_time);
+	WRITE_ONCE(cpuc->tot_sc_time, cpuc->tot_sc_time + sc_time);
+
+	taskc->acc_runtime += runtime;
+	taskc->svc_time += svc_time;
+	taskc->last_measured_clk = now;
+}
+
 static void update_stat_for_stopping(struct task_struct *p,
 				     struct task_ctx *taskc,
 				     struct cpu_ctx *cpuc)
 {
 	u64 now = scx_bpf_now();
-	u64 suspended_duration, task_runtime;
 
 	/*
-	 * Update task's runtime. When a task is scheduled consecutively
-	 * without ops.quiescent(), the task's runtime is accumulated for
-	 * statistics. Suppose a task is scheduled 2ms, 2ms, and 2ms with the
-	 * time slice exhausted. If 6ms of time slice was given in the first
-	 * place, the task will entirely consume the time slice. Hence, the
-	 * consecutive execution is accumulated and reflected in the
-	 * calculation of runtime statistics.
+	 * Account task runtime statistics first.
 	 */
-	suspended_duration = get_suspended_duration_and_reset(cpuc);
-	task_runtime = time_delta(now, taskc->last_running_clk + suspended_duration);
-	taskc->acc_runtime += task_runtime;
+	account_task_runtime(p, taskc, cpuc, now);
+
 	taskc->avg_runtime = calc_avg(taskc->avg_runtime, taskc->acc_runtime);
 	taskc->last_stopping_clk = now;
 
-	taskc->svc_time += task_runtime / p->scx.weight;
+	/*
+	 * Account for how much of the slice was used for this instance.
+	 */
+	if (is_monitored) {
+		taskc->last_slice_used = time_delta(now, taskc->last_running_clk);
+	}
 
 	/*
 	 * Reset waker's latency criticality here to limit the latency boost of
@@ -411,29 +476,34 @@ static void update_stat_for_stopping(struct task_struct *p,
 	 */
 	taskc->lat_cri_waker = 0;
 
-	/*
-	 * Increase total service time of this CPU.
-	 */
-	cpuc->tot_svc_time += taskc->svc_time;
-
 	/*
 	 * Update the current service time if necessary.
 	 */
 	if (READ_ONCE(cur_svc_time) < taskc->svc_time)
 		WRITE_ONCE(cur_svc_time, taskc->svc_time);
 
-	/*
-	 * Increase total scaled CPU time of this CPU,
-	 * which is capacity and frequency invariant.
-	 */
-	cpuc->tot_sc_time += scale_cap_freq(task_runtime, cpuc->cpu_id);
-
 	/*
 	 * Reset task's lock and futex boost count
 	 * for a lock holder to be boosted only once.
 	 */
 	reset_lock_futex_boost(taskc, cpuc);
-	taskc->lock_holder_xted = false;
+}
+
+static void update_stat_for_refill(struct task_struct *p,
+				   struct task_ctx *taskc,
+				   struct cpu_ctx *cpuc)
+{
+	u64 now = scx_bpf_now();
+
+	/*
+	 * Account task runtime statistics first.
+	 */
+	account_task_runtime(p, taskc, cpuc, now);
+
+	/*
+	 * We update avg_runtime here since it is used to boost time slice.
+	 */
+	taskc->avg_runtime = calc_avg(taskc->avg_runtime, taskc->acc_runtime);
 }
 
 s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
@@ -441,6 +511,7 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 {
 	bool found_idle = false;
 	struct task_ctx *taskc = get_task_ctx(p);
+	struct cpu_ctx *cpuc_cur = get_cpu_ctx();
 	struct cpu_ctx *cpuc;
 	u64 dsq_id;
 	s32 cpu_id;
@@ -448,19 +519,30 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 		.p = p,
 		.taskc = taskc,
 		.prev_cpu = prev_cpu,
+		.cpuc_cur = cpuc_cur,
 		.wake_flags = wake_flags,
 	};
 
-	if (!taskc)
+	if (!taskc || !cpuc_cur)
 		return prev_cpu;
-	taskc->wakeup_ft += !!(wake_flags & SCX_WAKE_SYNC);
+
+	if (wake_flags & SCX_WAKE_SYNC)
+		set_task_flag(taskc, LAVD_FLAG_IS_SYNC_WAKEUP);
+	else
+		reset_task_flag(taskc, LAVD_FLAG_IS_SYNC_WAKEUP);
 
 	/*
 	 * Find an idle cpu and reserve it since the task @p will run
-	 * on the idle cpu.
+	 * on the idle cpu. Even if there is no idle cpu, still respect
+	 * the chosen cpu.
 	 */
 	cpu_id = pick_idle_cpu(&ictx, &found_idle);
+	cpu_id = cpu_id >= 0 ? cpu_id : prev_cpu;
+	taskc->suggested_cpu_id = cpu_id;
+
 	if (found_idle) {
+		set_task_flag(taskc, LAVD_FLAG_IDLE_CPU_PICKED);
+
 		/*
 		 * If there is an idle cpu and its associated DSQ is empty,
 		 * disptach the task to the idle cpu right now.
@@ -468,22 +550,22 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 		cpuc = get_cpu_ctx_id(cpu_id);
 		if (!cpuc) {
 			scx_bpf_error("Failed to lookup cpu_ctx: %d", cpu_id);
-			return cpu_id;
+			goto out;
 		}
-		dsq_id = cpuc->cpdom_id;
+
+		dsq_id = cpdom_to_dsq(cpuc->cpdom_id);
 
 		if (!scx_bpf_dsq_nr_queued(dsq_id)) {
 			p->scx.dsq_vtime = calc_when_to_run(p, taskc);
-			p->scx.slice = calc_time_slice(taskc);
+			p->scx.slice = LAVD_SLICE_MAX_NS_DFL;
 			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, p->scx.slice, 0);
-			return cpu_id;
+			goto out;
 		}
+	} else {
+		reset_task_flag(taskc, LAVD_FLAG_IDLE_CPU_PICKED);
 	}
-
-	/*
-	 * Even if there is no idle cpu, still repect the chosen cpu.
-	 */
-	return cpu_id >= 0 ? cpu_id : prev_cpu;
+out:
+	return cpu_id;
 }
 
 static bool can_direct_dispatch(u64 dsq_id, s32 cpu, bool is_idle)
@@ -498,13 +580,17 @@ static bool can_direct_dispatch(u64 dsq_id, s32 cpu, bool is_idle)
 void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 {
 	struct task_ctx *taskc;
+	struct cpu_ctx *cpuc, *cpuc_cur;
 	s32 task_cpu, cpu = -ENOENT;
-	u64 dsq_id;
+	u64 cpdom_id;
 	bool is_idle = false;
 
 	taskc = get_task_ctx(p);
-	if (!taskc)
+	cpuc_cur = get_cpu_ctx();
+	if (!taskc || !cpuc_cur) {
+		scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
 		return;
+	}
 
 	/*
 	 * Calculate when a task can be scheduled for how long.
@@ -514,29 +600,65 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	 * timeslice, as the task hasn't yet run.
 	 */
 	if (!(enq_flags & SCX_ENQ_REENQ)) {
-		taskc->wakeup_ft += !!(enq_flags & SCX_ENQ_WAKEUP);
+		if (enq_flags & SCX_ENQ_WAKEUP)
+			set_task_flag(taskc, LAVD_FLAG_IS_WAKEUP);
+		else
+			reset_task_flag(taskc, LAVD_FLAG_IS_WAKEUP);
+
 		p->scx.dsq_vtime = calc_when_to_run(p, taskc);
 	}
-	p->scx.slice = calc_time_slice(taskc);
+	p->scx.slice = LAVD_SLICE_MAX_NS_DFL;
 
 	/*
 	 * Find a proper DSQ for the task, which is either the task's
 	 * associated compute domain or its alternative domain, or
 	 * the closest available domain from the previous domain.
+	 *
+	 * If the CPU is already picked at ops.select_cpu(),
+	 * let's use the chosen CPU.
 	 */
 	task_cpu = scx_bpf_task_cpu(p);
-	dsq_id = pick_proper_dsq(p, taskc, task_cpu, &cpu, &is_idle);
+	if (!__COMPAT_is_enq_cpu_selected(enq_flags)) {
+		cpdom_id = pick_proper_dsq(p, taskc, task_cpu, &cpu,
+					 &is_idle, cpuc_cur);
+		taskc->suggested_cpu_id = cpu;
+		cpuc = get_cpu_ctx_id(cpu);
+		if (!cpuc) {
+			scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
+			return;
+		}
+	} else {
+		cpu = scx_bpf_task_cpu(p);
+		cpuc = get_cpu_ctx_id(cpu);
+		if (!cpuc) {
+			scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
+			return;
+		}
+		cpdom_id = cpuc->cpdom_id;
+		is_idle = test_task_flag(taskc, LAVD_FLAG_IDLE_CPU_PICKED);
+		reset_task_flag(taskc, LAVD_FLAG_IDLE_CPU_PICKED);
+	}
+
+	/*
+	 * Increase the number of pinned tasks waiting for execution.
+	 */
+	if (is_pinned(p)) {
+		__sync_fetch_and_add(&cpuc->nr_pinned_tasks, 1);
+	}
 
 	/*
 	 * Enqueue the task to a DSQ. If it is safe to directly dispatch
 	 * to the local DSQ of the chosen CPU, do it. Otherwise, enqueue
 	 * to the chosen DSQ of the chosen domain.
 	 */
-	if (can_direct_dispatch(dsq_id, cpu, is_idle)) {
+	if (can_direct_dispatch(cpu_to_dsq(cpu), cpu, is_idle)) {
 		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, p->scx.slice,
 				   enq_flags);
+	} else if (per_cpu_dsq) {
+		scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu), p->scx.slice,
+					 p->scx.dsq_vtime, enq_flags);
 	} else {
-		scx_bpf_dsq_insert_vtime(p, dsq_id, p->scx.slice,
+		scx_bpf_dsq_insert_vtime(p, cpdom_to_dsq(cpdom_id), p->scx.slice,
 					 p->scx.dsq_vtime, enq_flags);
 	}
 
@@ -544,7 +666,7 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	 * If a new overflow CPU was assigned while finding a proper DSQ,
 	 * kick the new CPU and go.
 	 */
-	if (is_idle && cpu >= 0) {
+	if (is_idle) {
 		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 		return;
 	}
@@ -555,7 +677,7 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	 * from dsq_id. The kick will be done asynchronously.
 	 */
 	if (!no_preemption)
-		try_find_and_kick_victim_cpu(p, taskc, dsq_id);
+		try_find_and_kick_victim_cpu(p, taskc, cpu, cpdom_to_dsq(cpdom_id));
 }
 
 void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
@@ -564,41 +686,28 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 	struct task_ctx *taskc_prev = NULL;
 	struct bpf_cpumask *active, *ovrflw;
 	struct task_struct *p;
-	u64 dsq_id;
+	u32 dsq_type;
 	s32 new_cpu;
 	bool try_consume;
+	u64 dsq_ids[LAVD_DSQ_NR_TYPES];
+	int dsq_start = LAVD_DSQ_TYPE_CPDOM;
 
 	cpuc = get_cpu_ctx_id(cpu);
 	if (!cpuc) {
 		scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
 		return;
 	}
-	dsq_id = cpuc->cpdom_id;
+
+	dsq_ids[LAVD_DSQ_TYPE_CPU] = cpu_to_dsq(cpu);
+	dsq_ids[LAVD_DSQ_TYPE_CPDOM] = cpdom_to_dsq(cpuc->cpdom_id);
 
 	/*
 	 * If a task is holding a new lock, continue to execute it
 	 * to make system-wide forward progress.
 	 */
-	if (prev && (prev->scx.flags & SCX_TASK_QUEUED) && cpuc->lock_holder) {
-		taskc_prev = get_task_ctx(prev);
-		if (!taskc_prev) {
-			scx_bpf_error("Failed to look up task context");
-			return;
-		}
-
-		/*
-		 * Refill the time slice.
-		 */
-		prev->scx.slice = calc_time_slice(taskc_prev);
-
-		/*
-		 * Reset prev task's lock and futex boost count
-		 * for a lock holder to be boosted only once.
-		 */
-		reset_lock_futex_boost(taskc_prev, cpuc);
-		taskc_prev->lock_holder_xted = true;
-		return;
-	}
+	if (prev && (prev->scx.flags & SCX_TASK_QUEUED) &&
+	    is_lock_holder_running(cpuc))
+		goto consume_prev;
 
 	/*
 	 * If all CPUs are using, directly consume without checking CPU masks.
@@ -606,6 +715,13 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 	if (use_full_cpus())
 		goto consume_out;
 
+	/*
+	 * If there is something to run on a per-CPU DSQ,
+	 * directly consume without checking CPU masks.
+	 */
+	if (per_cpu_dsq && scx_bpf_dsq_nr_queued(dsq_ids[LAVD_DSQ_TYPE_CPU]))
+		goto consume_out;
+
 	/*
 	 * Prepare cpumasks.
 	 */
@@ -649,7 +765,8 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 		 * active or overflow set, extend the overflow set and go.
 		 */
 		taskc_prev = get_task_ctx(prev);
-		if (taskc_prev && taskc_prev->is_affinitized &&
+		if (taskc_prev &&
+		    test_task_flag(taskc_prev, LAVD_FLAG_IS_AFFINITIZED) &&
 		    bpf_cpumask_test_cpu(cpu, prev->cpus_ptr) &&
 		    !bpf_cpumask_intersects(cast_mask(active), prev->cpus_ptr) &&
 		    !bpf_cpumask_intersects(cast_mask(ovrflw), prev->cpus_ptr)) {
@@ -659,80 +776,92 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 		}
 	}
 
+	/*
+	 * Refactor this later to use some mapping to check if certain
+	 * dsq types are enabled.
+	 */
+	if (per_cpu_dsq)
+		dsq_start = LAVD_DSQ_TYPE_CPU;
 	/*
 	 * If this CPU is neither in active nor overflow CPUs,
 	 * try to find and run the task affinitized on this CPU.
 	 */
 	try_consume = false;
-	bpf_for_each(scx_dsq, p, dsq_id, 0) {
-		struct task_ctx *taskc;
+	bpf_for(dsq_type, dsq_start, LAVD_DSQ_NR_TYPES) {
+		u64 dsq_id = dsq_ids[dsq_type];
+		bpf_for_each(scx_dsq, p, dsq_id, 0) {
+			struct task_ctx *taskc;
 
-		/*
-		 * Note that this is a hack to bypass the restriction of the
-		 * current BPF not trusting the pointer p. Once the BPF
-		 * verifier gets smarter, we can remove bpf_task_from_pid().
-		 */
-		p = bpf_task_from_pid(p->pid);
-		if (!p)
+			/*
+			 * note that this is a hack to bypass the restriction of the
+			 * current bpf not trusting the pointer p. once the bpf
+			 * verifier gets smarter, we can remove bpf_task_from_pid().
+			 */
+			p = bpf_task_from_pid(p->pid);
+			if (!p)
+				break;
+
+			/*
+			 * if the task is pinned to this cpu,
+			 * extend the overflow set and go.
+			 * but not on this cpu, try another task.
+			 */
+			if (is_pinned(p)) {
+				new_cpu = scx_bpf_task_cpu(p);
+				if (new_cpu == cpu) {
+					bpf_cpumask_set_cpu(new_cpu, ovrflw);
+					bpf_task_release(p);
+					try_consume = true;
+					break;
+				}
+				if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
+					scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
+				bpf_task_release(p);
+				continue;
+			} else if (is_migration_disabled(p)) {
+				new_cpu = scx_bpf_task_cpu(p);
+				if (new_cpu == cpu) {
+					bpf_task_release(p);
+					try_consume = true;
+					break;
+				}
+				bpf_task_release(p);
+				continue;
+			}
+
+			/*
+			 * if the task can run on either active or overflow set,
+			 * try another task.
+			 */
+			taskc = get_task_ctx(p);
+			if(taskc &&
+			(!test_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED) ||
+			bpf_cpumask_intersects(cast_mask(active), p->cpus_ptr) ||
+			bpf_cpumask_intersects(cast_mask(ovrflw), p->cpus_ptr))) {
+				bpf_task_release(p);
+				continue;
+			}
+
+			/*
+			 * now, we know that the task cannot run on either active
+			 * or overflow set. then, let's consider to extend the
+			 * overflow set.
+			 */
+			new_cpu = find_cpu_in(p->cpus_ptr, cpuc);
+			if (new_cpu >= 0) {
+				if (new_cpu == cpu) {
+					bpf_cpumask_set_cpu(new_cpu, ovrflw);
+					bpf_task_release(p);
+					try_consume = true;
+					break;
+				}
+				else if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
+					scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
+			}
+			bpf_task_release(p);
+		}
+		if (try_consume)
 			break;
-
-		/*
-		 * If the task is pinned to this CPU,
-		 * extend the overflow set and go.
-		 * But not on this CPU, try another task.
-		 */
-		if (is_pinned(p)) {
-			new_cpu = scx_bpf_task_cpu(p);
-			if (new_cpu == cpu) {
-				bpf_cpumask_set_cpu(new_cpu, ovrflw);
-				bpf_task_release(p);
-				try_consume = true;
-				break;
-			}
-			if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
-				scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
-			bpf_task_release(p);
-			continue;
-		} else if (is_migration_disabled(p)) {
-			new_cpu = scx_bpf_task_cpu(p);
-			if (new_cpu == cpu) {
-				bpf_task_release(p);
-				try_consume = true;
-				break;
-			}
-			bpf_task_release(p);
-			continue;
-		}
-
-		/*
-		 * If the task can run on either active or overflow set,
-		 * try another task.
-		 */
-		taskc = get_task_ctx(p);
-		if(taskc && (!taskc->is_affinitized ||
-		   bpf_cpumask_intersects(cast_mask(active), p->cpus_ptr) ||
-		   bpf_cpumask_intersects(cast_mask(ovrflw), p->cpus_ptr))) {
-			bpf_task_release(p);
-			continue;
-		}
-
-		/*
-		 * Now, we know that the task cannot run on either active
-		 * or overflow set. Then, let's consider to extend the
-		 * overflow set.
-		 */
-		new_cpu = find_cpu_in(p->cpus_ptr, cpuc);
-		if (new_cpu >= 0) {
-			if (new_cpu == cpu) {
-				bpf_cpumask_set_cpu(new_cpu, ovrflw);
-				bpf_task_release(p);
-				try_consume = true;
-				break;
-			}
-			else if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
-				scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
-		}
-		bpf_task_release(p);
 	}
 
 unlock_out:
@@ -748,16 +877,40 @@ consume_out:
 	/*
 	 * Otherwise, consume a task.
 	 */
-	if (consume_task(dsq_id))
+	if (consume_task(dsq_ids[LAVD_DSQ_TYPE_CPU], dsq_ids[LAVD_DSQ_TYPE_CPDOM]))
 		return;
 
 	/*
 	 * If nothing to run, continue running the previous task.
 	 */
 	if (prev && prev->scx.flags & SCX_TASK_QUEUED) {
+consume_prev:
 		taskc_prev = taskc_prev ?: get_task_ctx(prev);
-		if (taskc_prev)
-			prev->scx.slice = calc_time_slice(taskc_prev);
+		if (taskc_prev) {
+			/*
+			 * Let's update stats first before calculating time slice.
+			 */
+			update_stat_for_refill(prev, taskc_prev, cpuc);
+
+			/*
+			 * Refill the time slice.
+			 */
+			prev->scx.slice = calc_time_slice(taskc_prev, cpuc);
+
+			/*
+			 * Reset prev task's lock and futex boost count
+			 * for a lock holder to be boosted only once.
+			 */
+			if (is_lock_holder_running(cpuc))
+				reset_lock_futex_boost(taskc_prev, cpuc);
+
+			/*
+			 * Task flags can be updated when calculating the time
+			 * slice (LAVD_FLAG_SLICE_BOOST), so let's update the
+			 * CPU's copy of the flag as well.
+			 */
+			cpuc->flags = taskc_prev->flags;
+		}
 	}
 }
 
@@ -816,15 +969,27 @@ void BPF_STRUCT_OPS(lavd_runnable, struct task_struct *p, u64 enq_flags)
 	 * Update wake frequency.
 	 */
 	now = scx_bpf_now();
-	interval = time_delta(now, waker_taskc->last_runnable_clk);
-	waker_taskc->wake_freq = calc_avg_freq(waker_taskc->wake_freq, interval);
-	waker_taskc->last_runnable_clk = now;
+	interval = time_delta(now, READ_ONCE(waker_taskc->last_runnable_clk));
+	if (interval >= LAVD_LC_WAKE_INTERVAL_MIN) {
+		WRITE_ONCE(waker_taskc->wake_freq,
+			   calc_avg_freq(waker_taskc->wake_freq, interval));
+		WRITE_ONCE(waker_taskc->last_runnable_clk, now);
+	}
 
 	/*
 	 * Propagate waker's latency criticality to wakee. Note that we pass
 	 * task's self latency criticality to limit the context into one hop.
 	 */
 	p_taskc->lat_cri_waker = waker_taskc->lat_cri;
+
+	/*
+	 * Collect additional information when the scheduler is monitored.
+	 */
+	if (is_monitored) {
+		p_taskc->waker_pid = waker->pid;
+		__builtin_memcpy_inline(p_taskc->waker_comm, waker->comm,
+					TASK_COMM_LEN);
+	}
 }
 
 void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
@@ -833,6 +998,67 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 	struct task_ctx *taskc;
 	u64 now = scx_bpf_now();
 
+	cpuc = get_cpu_ctx_task(p);
+	taskc = get_task_ctx(p);
+	if (!cpuc || !taskc) {
+		scx_bpf_error("Failed to lookup context for task %d", p->pid);
+		return;
+	}
+
+	/*
+	 * Increase the number of pinned tasks waiting for execution.
+	 */
+	if (is_pinned(p))
+		__sync_fetch_and_sub(&cpuc->nr_pinned_tasks, 1);
+
+	/*
+	 * If the sched_ext core directly dispatched a task, calculating the
+	 * task's deadline and time slice was also skipped. In this case, we
+	 * set the deadline to the current logical lock.
+	 *
+	 * Note that this is necessary when the kernel does not support
+	 * SCX_OPS_ENQ_MIGRATION_DISABLED or SCX_OPS_ENQ_MIGRATION_DISABLED
+	 * is not turned on.
+	 */
+	if (p->scx.slice == SCX_SLICE_DFL)
+		p->scx.dsq_vtime = READ_ONCE(cur_logical_clk);
+
+	/*
+	 * Calculate the task's time slice here,
+	 * as it depends on the system load.
+	 */
+	p->scx.slice = calc_time_slice(taskc, cpuc);
+
+	/*
+	 * Update the current logical clock.
+	 */
+	advance_cur_logical_clk(p);
+
+	/*
+	 * Update task statistics
+	 */
+	update_stat_for_running(p, taskc, cpuc, now);
+
+	/*
+	 * Calculate the task's CPU performance target and update if the new
+	 * target is higher than the current one. The CPU's performance target
+	 * urgently increases according to task's target but it decreases
+	 * gradually according to EWMA of past performance targets.
+	 */
+	update_cpuperf_target(cpuc);
+
+	/*
+	 * If there is a relevant introspection command with @p, process it.
+	 */
+	try_proc_introspec_cmd(p, taskc);
+}
+
+void BPF_STRUCT_OPS(lavd_tick, struct task_struct *p)
+{
+	struct cpu_ctx *cpuc;
+	struct task_ctx *taskc;
+	u64 now;
+
 	/*
 	 * Update task statistics
 	 */
@@ -843,31 +1069,17 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 		return;
 	}
 
-	update_stat_for_running(p, taskc, cpuc, now);
+	now = scx_bpf_now();
+	account_task_runtime(p, taskc, cpuc, now);
 
 	/*
-	 * Calculate the task's time slice.
+	 * If this task is slice-boosted and there is a pinned task that
+	 * must run on this, shrink its time slice to the regular one.
 	 */
-	p->scx.slice = calc_time_slice(taskc);
-
-	/*
-	 * Calculate the task's CPU performance target and update if the new
-	 * target is higher than the current one. The CPU's performance target
-	 * urgently increases according to task's target but it decreases
-	 * gradually according to EWMA of past performance targets.
-	 */
-	update_cpuperf_target(cpuc);
-
-	/*
-	 * Update running task's information for preemption
-	 */
-	cpuc->lat_cri = taskc->lat_cri;
-	cpuc->stopping_tm_est_ns = get_est_stopping_time(taskc, now);
-
-	/*
-	 * If there is a relevant introspection command with @p, process it.
-	 */
-	try_proc_introspec_cmd(p, taskc);
+	if (cpuc->nr_pinned_tasks &&
+	    test_cpu_flag(cpuc, LAVD_FLAG_SLICE_BOOST)) {
+		shrink_boosted_slice_at_tick(p, cpuc, now);
+	}
 }
 
 void BPF_STRUCT_OPS(lavd_stopping, struct task_struct *p, bool runnable)
@@ -894,15 +1106,13 @@ void BPF_STRUCT_OPS(lavd_quiescent, struct task_struct *p, u64 deq_flags)
 	struct task_ctx *taskc;
 	u64 now, interval;
 
-	/*
-	 * Substract task load from the current CPU's load.
-	 */
 	cpuc = get_cpu_ctx_task(p);
 	taskc = get_task_ctx(p);
 	if (!cpuc || !taskc) {
 		scx_bpf_error("Failed to lookup context for task %d", p->pid);
 		return;
 	}
+	cpuc->flags = 0;
 
 	/*
 	 * If a task @p is dequeued from a run queue for some other reason
@@ -917,8 +1127,10 @@ void BPF_STRUCT_OPS(lavd_quiescent, struct task_struct *p, u64 deq_flags)
 	 */
 	now = scx_bpf_now();
 	interval = time_delta(now, taskc->last_quiescent_clk);
-	taskc->wait_freq = calc_avg_freq(taskc->wait_freq, interval);
-	taskc->last_quiescent_clk = now;
+	if (interval > 0) {
+		taskc->wait_freq = calc_avg_freq(taskc->wait_freq, interval);
+		taskc->last_quiescent_clk = now;
+	}
 }
 
 static void cpu_ctx_init_online(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
@@ -933,9 +1145,11 @@ static void cpu_ctx_init_online(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
 unlock_out:
 	bpf_rcu_read_unlock();
 
+	cpuc->flags = 0;
 	cpuc->idle_start_clk = 0;
 	cpuc->lat_cri = 0;
-	cpuc->stopping_tm_est_ns = SCX_SLICE_INF;
+	cpuc->running_clk = 0;
+	cpuc->est_stopping_clk = SCX_SLICE_INF;
 	WRITE_ONCE(cpuc->online_clk, now);
 	barrier();
 
@@ -954,13 +1168,15 @@ static void cpu_ctx_init_offline(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
 unlock_out:
 	bpf_rcu_read_unlock();
 
+	cpuc->flags = 0;
 	cpuc->idle_start_clk = 0;
 	WRITE_ONCE(cpuc->offline_clk, now);
 	cpuc->is_online = false;
 	barrier();
 
 	cpuc->lat_cri = 0;
-	cpuc->stopping_tm_est_ns = SCX_SLICE_INF;
+	cpuc->running_clk = 0;
+	cpuc->est_stopping_clk = SCX_SLICE_INF;
 }
 
 void BPF_STRUCT_OPS(lavd_cpu_online, s32 cpu)
@@ -1077,7 +1293,10 @@ void BPF_STRUCT_OPS(lavd_set_cpumask, struct task_struct *p,
 		return;
 	}
 
-	taskc->is_affinitized = bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids;
+	if (bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids)
+		set_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
+	else
+		reset_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
 	set_on_core_type(taskc, cpumask);
 }
 
@@ -1122,6 +1341,7 @@ void BPF_STRUCT_OPS(lavd_cpu_release, s32 cpu,
 		scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
 		return;
 	}
+	cpuc->flags = 0;
 
 	/*
 	 * When a CPU is released to serve higher priority scheduler class,
@@ -1169,10 +1389,15 @@ static void init_task_ctx(struct task_struct *p, struct task_ctx *taskc)
 	u64 now = scx_bpf_now();
 
 	__builtin_memset(taskc, 0, sizeof(*taskc));
-	taskc->is_affinitized = bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids;
+	if (bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids)
+		set_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
+	else
+		reset_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
+	taskc->last_runnable_clk = now;
 	taskc->last_running_clk = now; /* for avg_runtime */
 	taskc->last_stopping_clk = now; /* for avg_runtime */
-	taskc->avg_runtime = slice_max_ns;
+	taskc->last_quiescent_clk = now;
+	taskc->avg_runtime = sys_stat.slice;
 	taskc->svc_time = sys_stat.avg_svc_time;
 
 	set_on_core_type(taskc, p->cpus_ptr);
@@ -1232,7 +1457,7 @@ static s32 init_cpdoms(u64 now)
 		/*
 		 * Create an associated DSQ on its associated NUMA domain.
 		 */
-		err = scx_bpf_create_dsq(cpdomc->id, cpdomc->node_id);
+		err = scx_bpf_create_dsq(cpdom_to_dsq(cpdomc->id), cpdomc->node_id);
 		if (err) {
 			scx_bpf_error("Failed to create a DSQ for cpdom %llu on NUMA node %d",
 				      cpdomc->id, cpdomc->node_id);
@@ -1378,7 +1603,8 @@ static s32 init_per_cpu_ctx(u64 now)
 		cpuc->cpu_id = cpu;
 		cpuc->idle_start_clk = 0;
 		cpuc->lat_cri = 0;
-		cpuc->stopping_tm_est_ns = SCX_SLICE_INF;
+		cpuc->running_clk = 0;
+		cpuc->est_stopping_clk = SCX_SLICE_INF;
 		cpuc->online_clk = now;
 		cpuc->offline_clk = now;
 		cpuc->cpu_release_clk = now;
@@ -1477,6 +1703,41 @@ unlock_out:
 	return err;
 }
 
+
+static int init_per_cpu_dsqs(void)
+{
+	struct cpdom_ctx *cpdomc;
+	struct cpu_ctx *cpuc;
+	int cpu, err = 0;
+
+	bpf_for(cpu, 0, nr_cpu_ids) {
+		/*
+		 * Create Per-CPU DSQs on its associated NUMA domain.
+		 */
+		cpuc = get_cpu_ctx_id(cpu);
+		if (!cpuc) {
+			scx_bpf_error("Failed to lookup cpu_ctx: %d", cpu);
+			return -ESRCH;
+		}
+
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpuc->cpdom_id]);
+		if (!cpdomc) {
+			scx_bpf_error("Failed to lookup cpdom_ctx for %hhu", cpuc->cpdom_id);
+			return -ESRCH;
+		}
+
+		err = scx_bpf_create_dsq(cpu_to_dsq(cpu), cpdomc->node_id);
+		if (err) {
+			scx_bpf_error("Failed to create a DSQ for cpu %d on NUMA node %d",
+				      cpu, cpdomc->node_id);
+			return err;
+		}
+	}
+
+	return 0;
+}
+
+
 s32 BPF_STRUCT_OPS_SLEEPABLE(lavd_init)
 {
 	u64 now = scx_bpf_now();
@@ -1506,6 +1767,15 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(lavd_init)
 	if (err)
 		return err;
 
+	/*
+	 * Initialize per-CPU DSQs.
+	 */
+	if (per_cpu_dsq) {
+		err = init_per_cpu_dsqs();
+		if (err)
+			return err;
+	}
+
 	/*
 	 * Initialize the last update clock and the update timer to track
 	 * system-wide CPU load.
@@ -1539,6 +1809,7 @@ SCX_OPS_DEFINE(lavd_ops,
 	       .dispatch		= (void *)lavd_dispatch,
 	       .runnable		= (void *)lavd_runnable,
 	       .running			= (void *)lavd_running,
+	       .tick			= (void *)lavd_tick,
 	       .stopping		= (void *)lavd_stopping,
 	       .quiescent		= (void *)lavd_quiescent,
 	       .cpu_online		= (void *)lavd_cpu_online,
