@@ -90,30 +90,41 @@ volatile u64		performance_mode_ns;
 volatile u64		balanced_mode_ns;
 volatile u64		powersave_mode_ns;
 
+static void reset_suspended_duration(struct cpu_ctx *cpuc)
+{
+	if (cpuc->online_clk > cpuc->offline_clk)
+		cpuc->offline_clk = cpuc->online_clk;
+}
+
+static u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
+{
+	/*
+	 * When a system is suspended, a task is also suspended in a running
+	 * stat on the CPU. Hence, we subtract the suspended duration when it
+	 * resumes.
+	 */
+	u64 duration = 0;
+
+	if (cpuc->online_clk > cpuc->offline_clk) {
+		duration = time_delta(cpuc->online_clk, cpuc->offline_clk);
+		/*
+		 * Once calculated, reset the duration to zero.
+		 */
+		cpuc->offline_clk = cpuc->online_clk;
+	}
+
+	return duration;
+}
+
 static bool is_perf_cri(struct task_ctx *taskc)
 {
 	if (!have_little_core)
 		return true;
 
-	if (READ_ONCE(taskc->on_big) && READ_ONCE(taskc->on_little))
+	if (test_task_flag(taskc, LAVD_FLAG_ON_BIG | LAVD_FLAG_ON_LITTLE))
 		return taskc->perf_cri >= sys_stat.thr_perf_cri;
-	return READ_ONCE(taskc->on_big);
-}
 
-static bool clear_cpu_periodically(u32 cpu, struct bpf_cpumask *cpumask)
-{
-	u32 clear;
-
-	/*
-	 * If the CPU is on, we clear the bit once every four times
-	 * (LAVD_CC_CPU_PIN_INTERVAL_DIV). Hence, the bit will be
-	 * probabilistically cleared once every 100 msec (4 * 25 msec).
-	 */
-	clear = !(bpf_get_prandom_u32() % LAVD_CC_CPU_PIN_INTERVAL_DIV);
-	if (clear)
-		bpf_cpumask_clear_cpu(cpu, cpumask);
-
-	return clear;
+	return test_task_flag(taskc, LAVD_FLAG_ON_BIG);
 }
 
 static const volatile u16 *get_cpu_order(void)
@@ -227,9 +238,9 @@ static void do_core_compaction(void)
 	struct cpu_ctx *cpuc;
 	struct bpf_cpumask *active, *ovrflw;
 	struct cpdom_ctx *cpdomc;
-	int nr_active, nr_active_old, cpu, i;
+	int nr_active, cpu, i;
 	u32 sum_capacity = 0, big_capacity = 0, nr_active_cpdoms = 0;
-	bool clear;
+	bool need_kick;
 	u64 cpdom_id;
 
 	bpf_rcu_read_lock();
@@ -250,7 +261,6 @@ static void do_core_compaction(void)
 	 * active CPUs. Finally, obtain the CPU order list based on the current
 	 * load.
 	 */
-	nr_active_old = sys_stat.nr_active;
 	nr_active = calc_nr_active_cpus();
 	cpu_order = get_cpu_order();
 
@@ -275,10 +285,11 @@ static void do_core_compaction(void)
 		/*
 		 * Assign an online cpu to active and overflow cpumasks
 		 */
+		need_kick = false;
 		if (i < nr_active) {
 			bpf_cpumask_set_cpu(cpu, active);
 			bpf_cpumask_clear_cpu(cpu, ovrflw);
-			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+			need_kick = true;
 
 			/*
 			 * Accumulate the capacity of active CPUs and
@@ -296,29 +307,44 @@ static void do_core_compaction(void)
 			sum_capacity += cpuc->capacity;
 			if (cpuc->big_core)
 				big_capacity += cpuc->capacity;
-		} else if (i < nr_active_old) {
-			bpf_cpumask_clear_cpu(cpu, active);
-			bpf_cpumask_clear_cpu(cpu, ovrflw);
 		} else {
-			/*
-			 * This is the case when a CPU belongs to the
-			 * overflow set even though that CPU was not an
-			 * overflow set initially. This can happen only
-			 * when a pinned userspace task ran on this
-			 * CPU. In this case, we keep the CPU in an
-			 * overflow set since the CPU will be used
-			 * anyway for the task. This will promote equal
-			 * use of all used CPUs, lowering the energy
-			 * consumption by avoiding a few CPUs being
-			 * turbo-boosted. Hence, we do not clear the
-			 * overflow cpumask here for a while,
-			 * approximately for LAVD_CC_CPU_PIN_INTERVAL.
-			 */
 			bpf_cpumask_clear_cpu(cpu, active);
-			clear = clear_cpu_periodically(cpu, ovrflw);
-			if (!clear)
-				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+
+			if (cpuc->nr_pinned_tasks || (per_cpu_dsq &&
+			    scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu)))) {
+				/*
+				 * If there is something to run on this CPU,
+				 * add this CPU{ to the overflow set.
+				 */
+				bpf_cpumask_set_cpu(cpu, ovrflw);
+				need_kick = true;
+			} else if ((bpf_get_prandom_u32() %
+				    LAVD_CC_CPU_PIN_INTERVAL_DIV)) {
+				/*
+				 * This is the case when a CPU belongs to the
+				 * overflow set even though that CPU was not an
+				 * overflow set initially. This can happen only
+				 * when a pinned userspace task ran on this
+				 * CPU. In this case, we keep the CPU in an
+				 * overflow set since the CPU will be used
+				 * anyway for the task. This will promote equal
+				 * use of all used CPUs, lowering the energy
+				 * consumption by avoiding a few CPUs being
+				 * turbo-boosted. Hence, we do not clear the
+				 * overflow cpumask here for a while,
+				 * approximately for LAVD_CC_CPU_PIN_INTERVAL.
+				 */
+				bpf_cpumask_clear_cpu(cpu, ovrflw);
+			} else if (bpf_cpumask_test_cpu(cpu, cast_mask(ovrflw))) {
+				need_kick = true;
+			}
 		}
+
+		/*
+		 * When the CPU is in either an active or overflow set, kick it.
+		 */
+		if (need_kick)
+			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 	}
 
 	cur_big_core_scale = (big_capacity << LAVD_SHIFT) / sum_capacity;
@@ -546,7 +572,7 @@ static int reinit_active_cpumask_for_performance(void)
 	struct bpf_cpumask *active, *ovrflw;
 	const struct cpumask *online_cpumask;
 	struct cpdom_ctx *cpdomc;
-	u64 dsq_id;
+	u64 cpdom_id;
 	u32 nr_active_cpdoms = 0;
 	int cpu, err = 0;
 
@@ -625,11 +651,11 @@ static int reinit_active_cpumask_for_performance(void)
 	/*
 	 * Update nr_active_cpus, cap_sum_active_cpus, and pco_idx.
 	 */
-	bpf_for(dsq_id, 0, nr_cpdoms) {
-		if (dsq_id >= LAVD_CPDOM_MAX_NR)
+	bpf_for(cpdom_id, 0, nr_cpdoms) {
+		if (cpdom_id >= LAVD_CPDOM_MAX_NR)
 			break;
 
-		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpdom_id]);
 		WRITE_ONCE(cpdomc->nr_active_cpus, cpdomc->nr_acpus_temp);
 		WRITE_ONCE(cpdomc->nr_acpus_temp, 0);
 		WRITE_ONCE(cpdomc->cap_sum_active_cpus, cpdomc->cap_sum_temp);
