@@ -59,25 +59,40 @@ static void init_sys_stat_ctx(struct sys_stat_ctx *c)
 	c->min_perf_cri = LAVD_SCALE;
 	c->now = scx_bpf_now();
 	c->duration = time_delta(c->now, sys_stat.last_update_clk);
-	sys_stat.last_update_clk = c->now;
+	WRITE_ONCE(sys_stat.last_update_clk, c->now);
 }
 
 static void collect_sys_stat(struct sys_stat_ctx *c)
 {
 	struct cpdom_ctx *cpdomc;
-	u64 dsq_id, cpuc_tot_sc_time, compute;
+	u64 cpdom_id, cpuc_tot_sc_time, compute;
 	int cpu;
 
 	/*
 	 * Collect statistics for each compute domain.
 	 */
-	bpf_for(dsq_id, 0, nr_cpdoms) {
-		if (dsq_id >= LAVD_CPDOM_MAX_NR)
+	bpf_for(cpdom_id, 0, nr_cpdoms) {
+		int i, j;
+		if (cpdom_id >= LAVD_CPDOM_MAX_NR)
 			break;
 
-		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpdom_id]);
 		cpdomc->cur_util_sum = 0;
-		cpdomc->nr_queued_task = scx_bpf_dsq_nr_queued(dsq_id);
+		cpdomc->nr_queued_task = scx_bpf_dsq_nr_queued(cpdom_to_dsq(cpdom_id));
+		if (per_cpu_dsq) {
+			bpf_for(i, 0, LAVD_CPU_ID_MAX/64) {
+				u64 cpumask = cpdomc->__cpumask[i];
+				bpf_for(j, 0, 64) {
+					if (cpumask & 0x1LLU << j) {
+						cpu = (i * 64) + j;
+						if (cpu >= nr_cpu_ids)
+							break;
+						cpdomc->nr_queued_task += scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu));
+					}
+				}
+			}
+		}
+
 		c->nr_queued_task += cpdomc->nr_queued_task;
 	}
 
@@ -91,6 +106,17 @@ static void collect_sys_stat(struct sys_stat_ctx *c)
 			break;
 		}
 
+		/*
+		 * When pinned tasks are waiting to run on this CPU
+		 * or a system is overloaded (so the slice cannot be boosted
+		 * or there are pending tasks to run), shrink the time slice
+		 * of slice-boosted tasks.
+		 */
+		if (cpuc->nr_pinned_tasks || !can_boost_slice() ||
+		    scx_bpf_dsq_nr_queued(SCX_DSQ_LOCAL_ON | cpuc->cpu_id)) {
+			shrink_boosted_slice_remote(cpuc, c->now);
+		}
+
 		/*
 		 * Accumulate cpus' loads.
 		 */
@@ -208,15 +234,6 @@ static void collect_sys_stat(struct sys_stat_ctx *c)
 	}
 }
 
-static u64 clamp_time_slice_ns(u64 slice)
-{
-	if (slice < slice_min_ns)
-		slice = slice_min_ns;
-	else if (slice > slice_max_ns)
-		slice = slice_max_ns;
-	return slice;
-}
-
 static void calc_sys_stat(struct sys_stat_ctx *c)
 {
 	static int cnt = 0;
@@ -341,7 +358,7 @@ static void calc_sys_stat(struct sys_stat_ctx *c)
 
 static void calc_sys_time_slice(void)
 {
-	u64 nr_queued, slice;
+	u64 nr_q, slice;
 
 	/*
 	 * Given the updated state, recalculate the time slice for the next
@@ -349,9 +366,13 @@ static void calc_sys_time_slice(void)
 	 * runnable tasks at least once within a targeted latency using the
 	 * active CPUs.
 	 */
-	nr_queued = sys_stat.nr_queued_task + 1;
-	slice = (LAVD_TARGETED_LATENCY_NS * sys_stat.nr_active) / nr_queued;
-	slice = clamp_time_slice_ns(slice);
+	nr_q = sys_stat.nr_queued_task;
+	if (nr_q > 0) {
+		slice = (LAVD_TARGETED_LATENCY_NS * sys_stat.nr_active) / nr_q;
+		slice = clamp(slice, slice_min_ns, slice_max_ns);
+	} else {
+		slice = slice_max_ns;
+	}
 	sys_stat.slice = calc_avg(sys_stat.slice, slice);
 }
 
@@ -419,18 +440,18 @@ static s32 init_sys_stat(u64 now)
 {
 	struct cpdom_ctx *cpdomc;
 	struct bpf_timer *timer;
-	u64 dsq_id;
+	u64 cpdom_id;
 	u32 key = 0;
 	int err;
 
 	sys_stat.last_update_clk = now;
 	sys_stat.nr_active = nr_cpus_onln;
-	sys_stat.slice = LAVD_SLICE_MAX_NS_DFL;
-	bpf_for(dsq_id, 0, nr_cpdoms) {
-		if (dsq_id >= LAVD_CPDOM_MAX_NR)
+	sys_stat.slice = slice_max_ns;
+	bpf_for(cpdom_id, 0, nr_cpdoms) {
+		if (cpdom_id >= LAVD_CPDOM_MAX_NR)
 			break;
 
-		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpdom_id]);
 		if (cpdomc->nr_active_cpus)
 			sys_stat.nr_active_cpdoms++;
 	}
