@@ -26,12 +26,15 @@ struct sys_stat	sys_stat;
 /*
  * Options
  */
-volatile bool		no_preemption;
-volatile bool		no_wake_sync;
-volatile bool		no_core_compaction;
-volatile bool		no_freq_scaling;
 volatile bool		is_powersave_mode;
 volatile bool		reinit_cpumask_for_performance;
+volatile bool		no_preemption;
+volatile bool		no_core_compaction;
+volatile bool		no_freq_scaling;
+
+const volatile bool	no_wake_sync;
+const volatile bool	no_slice_boost;
+const volatile bool	per_cpu_dsq;
 const volatile bool	is_autopilot_on;
 const volatile u8	verbose;
 const volatile u8	preempt_shift;
@@ -84,6 +87,10 @@ struct {
 #define max(X, Y) (((X) < (Y)) ? (Y) : (X))
 #endif
 
+#ifndef clamp
+#define clamp(val, lo, hi) min(max(val, lo), hi)
+#endif
+
 static struct task_ctx *get_task_ctx(struct task_struct *p)
 {
 	return bpf_task_storage_get(&task_ctx_stor, p, 0, 0);
@@ -163,24 +170,49 @@ static bool is_pinned(const struct task_struct *p)
 	return p->nr_cpus_allowed == 1;
 }
 
+static inline bool test_task_flag(struct task_ctx *taskc, u64 flag)
+{
+	return (taskc->flags & flag) == flag;
+}
+
+static inline void set_task_flag(struct task_ctx *taskc, u64 flag)
+{
+	taskc->flags |= flag;
+}
+
+static inline void reset_task_flag(struct task_ctx *taskc, u64 flag)
+{
+	taskc->flags &= ~flag;
+}
+
+static inline bool test_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
+{
+	return (cpuc->flags & flag) == flag;
+}
+
+static inline void set_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
+{
+	cpuc->flags |= flag;
+}
+
+static inline void reset_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
+{
+	cpuc->flags &= ~flag;
+}
+
 static bool is_lat_cri(struct task_ctx *taskc)
 {
 	return taskc->lat_cri >= sys_stat.avg_lat_cri;
 }
 
-static bool is_greedy(struct task_ctx *taskc)
-{
-	return taskc->is_greedy;
-}
-
-static bool is_eligible(struct task_ctx *taskc)
-{
-	return !is_greedy(taskc);
-}
-
 static bool is_lock_holder(struct task_ctx *taskc)
 {
-	return taskc->futex_boost;
+	return test_task_flag(taskc, LAVD_FLAG_FUTEX_BOOST);
+}
+
+static bool is_lock_holder_running(struct cpu_ctx *cpuc)
+{
+	return test_cpu_flag(cpuc, LAVD_FLAG_FUTEX_BOOST);
 }
 
 static bool have_scheduled(struct task_ctx *taskc)
@@ -189,7 +221,12 @@ static bool have_scheduled(struct task_ctx *taskc)
 	 * If task's time slice hasn't been updated, that means the task has
 	 * been scheduled by this scheduler.
 	 */
-	return taskc->slice_ns != 0;
+	return taskc->slice != 0;
+}
+
+static bool can_boost_slice(void)
+{
+	return sys_stat.nr_queued_task <= sys_stat.nr_active;
 }
 
 static u16 get_nice_prio(struct task_struct *p)
@@ -242,17 +279,29 @@ static void set_on_core_type(struct task_ctx *taskc,
 			break;
 	}
 
-	WRITE_ONCE(taskc->on_big, on_big);
-	WRITE_ONCE(taskc->on_little, on_little);
+	if (on_big)
+		set_task_flag(taskc, LAVD_FLAG_ON_BIG);
+	else
+		reset_task_flag(taskc, LAVD_FLAG_ON_BIG);
+
+	if (on_little)
+		set_task_flag(taskc, LAVD_FLAG_ON_LITTLE);
+	else
+		reset_task_flag(taskc, LAVD_FLAG_ON_LITTLE);
 }
 
-static bool prob_x_out_of_y(u32 x, u32 y)
+bool __attribute__ ((noinline)) prob_x_out_of_y(u32 x, u32 y)
 {
+	u32 r;
+
+	if (x >= y)
+		return true;
+
 	/*
 	 * [0, r, y)
 	 *  ---- x?
 	 */
-	u32 r = bpf_get_prandom_u32() % y;
+	r = bpf_get_prandom_u32() % y;
 	return r < x;
 }
 
