@@ -12,6 +12,7 @@ pub mod bpf_intf;
 pub use bpf_intf::*;
 
 mod cpu_order;
+use scx_utils::init_libbpf_logging;
 mod stats;
 use std::ffi::c_int;
 use std::ffi::CStr;
@@ -36,6 +37,7 @@ use crossbeam::channel::RecvTimeoutError;
 use crossbeam::channel::Sender;
 use crossbeam::channel::TrySendError;
 use libbpf_rs::OpenObject;
+use libbpf_rs::PrintLevel;
 use libbpf_rs::ProgramInput;
 use libc::c_char;
 use log::debug;
@@ -145,6 +147,19 @@ struct Opts {
     #[clap(long = "no-wake-sync", action = clap::ArgAction::SetTrue)]
     no_wake_sync: bool,
 
+    /// Disable dynamic slice boost for long-running tasks.
+    #[clap(long = "no-slice-boost", action = clap::ArgAction::SetTrue)]
+    no_slice_boost: bool,
+
+    /// Enables DSQs per CPU, this enables task queuing and dispatching
+    /// from CPU specific DSQs. This generally increases L1/L2 cache
+    /// locality for tasks and lowers lock contention compared to shared DSQs,
+    /// but at the cost of higher load balancing complexity. This is a
+    /// highly experimental feature.
+    #[clap(long = "per-cpu-dsq", action = clap::ArgAction::SetTrue)]
+    per_cpu_dsq: bool,
+
+    ///
     /// Disable core compaction so the scheduler uses all the online CPUs.
     /// The core compaction attempts to minimize the number of actively used
     /// CPUs for unaffinitized tasks, respecting the CPU preference order.
@@ -330,6 +345,8 @@ impl<'a> Scheduler<'a> {
         // Open the BPF prog first for verification.
         let mut skel_builder = BpfSkelBuilder::default();
         skel_builder.obj_builder.debug(opts.verbose > 0);
+        init_libbpf_logging(Some(PrintLevel::Debug));
+
         let mut skel = scx_ops_open!(skel_builder, open_object, lavd_ops)?;
 
         // Enable futex tracing using ftrace if available. If the ftrace is not
@@ -422,9 +439,9 @@ impl<'a> Scheduler<'a> {
 
         // Initialize CPU capacity.
         for cpu in order.cpuids.iter() {
-            skel.maps.rodata_data.cpu_capacity[cpu.cpu_adx] = cpu.cpu_cap as u16;
-            skel.maps.rodata_data.cpu_big[cpu.cpu_adx] = cpu.big_core as u8;
-            skel.maps.rodata_data.cpu_turbo[cpu.cpu_adx] = cpu.turbo_core as u8;
+            skel.maps.rodata_data.as_mut().unwrap().cpu_capacity[cpu.cpu_adx] = cpu.cpu_cap as u16;
+            skel.maps.rodata_data.as_mut().unwrap().cpu_big[cpu.cpu_adx] = cpu.big_core as u8;
+            skel.maps.rodata_data.as_mut().unwrap().cpu_turbo[cpu.cpu_adx] = cpu.turbo_core as u8;
         }
 
         // Initialize performance vs. CPU order table.
@@ -433,7 +450,7 @@ impl<'a> Scheduler<'a> {
             panic!("Generated performance vs. CPU order stats are too complex ({nr_pco_states}) to handle");
         }
 
-        skel.maps.rodata_data.nr_pco_states = nr_pco_states;
+        skel.maps.rodata_data.as_mut().unwrap().nr_pco_states = nr_pco_states;
         for (i, (_, pco)) in order.perf_cpu_order.iter().enumerate() {
             Self::init_pco_tuple(skel, i, &pco);
             info!("{:#}", pco);
@@ -450,31 +467,33 @@ impl<'a> Scheduler<'a> {
         let cpus_ovflw = pco.cpus_ovflw.borrow();
         let pco_nr_primary = cpus_perf.len();
 
-        skel.maps.rodata_data.pco_bounds[i] = pco.perf_cap as u32;
-        skel.maps.rodata_data.pco_nr_primary[i] = pco_nr_primary as u16;
+        skel.maps.rodata_data.as_mut().unwrap().pco_bounds[i] = pco.perf_cap as u32;
+        skel.maps.rodata_data.as_mut().unwrap().pco_nr_primary[i] = pco_nr_primary as u16;
 
         for (j, &cpu_adx) in cpus_perf.iter().enumerate() {
-            skel.maps.rodata_data.pco_table[i][j] = cpu_adx as u16;
+            skel.maps.rodata_data.as_mut().unwrap().pco_table[i][j] = cpu_adx as u16;
         }
 
         for (j, &cpu_adx) in cpus_ovflw.iter().enumerate() {
             let k = j + pco_nr_primary;
-            skel.maps.rodata_data.pco_table[i][k] = cpu_adx as u16;
+            skel.maps.rodata_data.as_mut().unwrap().pco_table[i][k] = cpu_adx as u16;
         }
     }
 
     fn init_cpdoms(skel: &mut OpenBpfSkel, order: &CpuOrder) {
         // Initialize compute domain contexts
         for (k, v) in order.cpdom_map.iter() {
-            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].id = v.cpdom_id as u64;
-            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].alt_id = v.cpdom_alt_id.get() as u64;
-            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].node_id = k.node_adx as u8;
-            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].is_big = k.is_big as u8;
-            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].is_valid = 1;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].id = v.cpdom_id as u64;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].alt_id =
+                v.cpdom_alt_id.get() as u64;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].node_id = k.node_adx as u8;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].is_big = k.is_big as u8;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].is_valid = 1;
             for cpu_id in v.cpu_ids.iter() {
                 let i = cpu_id / 64;
                 let j = cpu_id % 64;
-                skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].__cpumask[i] |= 0x01 << j;
+                skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].__cpumask[i] |=
+                    0x01 << j;
             }
 
             if v.neighbor_map.borrow().iter().len() > LAVD_CPDOM_MAX_DIST as usize {
@@ -486,28 +505,34 @@ impl<'a> Scheduler<'a> {
                 if nr_neighbors > LAVD_CPDOM_MAX_NR as u8 {
                     panic!("The processor topology is too complex to handle in BPF.");
                 }
-                skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].nr_neighbors[k] = nr_neighbors;
+                skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].nr_neighbors[k] =
+                    nr_neighbors;
                 for n in neighbors.borrow().iter() {
-                    skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].neighbor_bits[k] |= 0x1 << n;
+                    skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].neighbor_bits[k] |=
+                        0x1 << n;
                 }
             }
         }
     }
 
     fn init_globals(skel: &mut OpenBpfSkel, opts: &Opts, order: &CpuOrder) {
-        skel.maps.bss_data.no_preemption = opts.no_preemption;
-        skel.maps.bss_data.no_wake_sync = opts.no_wake_sync;
-        skel.maps.bss_data.no_core_compaction = opts.no_core_compaction;
-        skel.maps.bss_data.no_freq_scaling = opts.no_freq_scaling;
-        skel.maps.bss_data.is_powersave_mode = opts.powersave;
-        skel.maps.rodata_data.nr_cpu_ids = *NR_CPU_IDS as u64;
-        skel.maps.rodata_data.is_smt_active = order.smt_enabled;
-        skel.maps.rodata_data.is_autopilot_on = opts.autopilot;
-        skel.maps.rodata_data.verbose = opts.verbose;
-        skel.maps.rodata_data.slice_max_ns = opts.slice_max_us * 1000;
-        skel.maps.rodata_data.slice_min_ns = opts.slice_min_us * 1000;
-        skel.maps.rodata_data.preempt_shift = opts.preempt_shift;
-        skel.maps.rodata_data.no_use_em = opts.no_use_em as u8;
+        let bss_data = skel.maps.bss_data.as_mut().unwrap();
+        bss_data.no_preemption = opts.no_preemption;
+        bss_data.no_core_compaction = opts.no_core_compaction;
+        bss_data.no_freq_scaling = opts.no_freq_scaling;
+        bss_data.is_powersave_mode = opts.powersave;
+        let rodata = skel.maps.rodata_data.as_mut().unwrap();
+        rodata.nr_cpu_ids = *NR_CPU_IDS as u64;
+        rodata.is_smt_active = order.smt_enabled;
+        rodata.is_autopilot_on = opts.autopilot;
+        rodata.verbose = opts.verbose;
+        rodata.slice_max_ns = opts.slice_max_us * 1000;
+        rodata.slice_min_ns = opts.slice_min_us * 1000;
+        rodata.preempt_shift = opts.preempt_shift;
+        rodata.no_use_em = opts.no_use_em as u8;
+        rodata.no_wake_sync = opts.no_wake_sync;
+        rodata.no_slice_boost = opts.no_slice_boost;
+        rodata.per_cpu_dsq = opts.per_cpu_dsq;
 
         skel.struct_ops.lavd_ops_mut().flags = *compat::SCX_OPS_ENQ_EXITING
             | *compat::SCX_OPS_ENQ_LAST
@@ -539,6 +564,10 @@ impl<'a> Scheduler<'a> {
         let c_tx_cm_str: &CStr = unsafe { CStr::from_ptr(c_tx_cm) };
         let tx_comm: &str = c_tx_cm_str.to_str().unwrap();
 
+        let c_waker_cm: *const c_char = (&tc.waker_comm as *const [c_char; 17]) as *const c_char;
+        let c_waker_cm_str: &CStr = unsafe { CStr::from_ptr(c_waker_cm) };
+        let waker_comm: &str = c_waker_cm_str.to_str().unwrap();
+
         let c_tx_st: *const c_char = (&tx.stat as *const [c_char; 5]) as *const c_char;
         let c_tx_st_str: &CStr = unsafe { CStr::from_ptr(c_tx_st) };
         let tx_stat: &str = c_tx_st_str.to_str().unwrap();
@@ -548,11 +577,17 @@ impl<'a> Scheduler<'a> {
             pid: tx.pid,
             comm: tx_comm.into(),
             stat: tx_stat.into(),
-            cpu_id: tx.cpu_id,
-            slice_ns: tc.slice_ns,
+            cpu_id: tc.cpu_id,
+            prev_cpu_id: tc.prev_cpu_id,
+            suggested_cpu_id: tc.suggested_cpu_id,
+            waker_pid: tc.waker_pid,
+            waker_comm: waker_comm.into(),
+            slice: tc.slice,
             lat_cri: tc.lat_cri,
             avg_lat_cri: tx.avg_lat_cri,
             static_prio: tx.static_prio,
+            rerunnable_interval: tx.rerunnable_interval,
+            resched_interval: tc.resched_interval,
             run_freq: tc.run_freq,
             avg_runtime: tc.avg_runtime,
             wait_freq: tc.wait_freq,
@@ -563,6 +598,9 @@ impl<'a> Scheduler<'a> {
             cpu_util: tx.cpu_util,
             cpu_sutil: tx.cpu_sutil,
             nr_active: tx.nr_active,
+            dsq_id: tx.dsq_id,
+            dsq_consume_lat: tx.dsq_consume_lat,
+            slice_used: tc.last_slice_used,
         }) {
             Ok(()) | Err(TrySendError::Full(_)) => 0,
             Err(e) => panic!("failed to send on intrspc_tx ({})", e),
@@ -570,12 +608,15 @@ impl<'a> Scheduler<'a> {
     }
 
     fn prep_introspec(&mut self) {
-        self.skel.maps.bss_data.intrspc.cmd = self.intrspc.cmd;
-        self.skel.maps.bss_data.intrspc.arg = self.intrspc.arg;
+        if !self.skel.maps.bss_data.as_ref().unwrap().is_monitored {
+            self.skel.maps.bss_data.as_mut().unwrap().is_monitored = true;
+        }
+        self.skel.maps.bss_data.as_mut().unwrap().intrspc.cmd = self.intrspc.cmd;
+        self.skel.maps.bss_data.as_mut().unwrap().intrspc.arg = self.intrspc.arg;
     }
 
     fn cleanup_introspec(&mut self) {
-        self.skel.maps.bss_data.intrspc.cmd = LAVD_CMD_NOP;
+        self.skel.maps.bss_data.as_mut().unwrap().intrspc.cmd = LAVD_CMD_NOP;
     }
 
     fn get_pc(x: u64, y: u64) -> f64 {
@@ -604,7 +645,7 @@ impl<'a> Scheduler<'a> {
                 }
                 self.mseq_id += 1;
 
-                let bss_data = &self.skel.maps.bss_data;
+                let bss_data = self.skel.maps.bss_data.as_ref().unwrap();
                 let st = bss_data.sys_stat;
 
                 let mseq = self.mseq_id;
@@ -674,6 +715,12 @@ impl<'a> Scheduler<'a> {
         })
     }
 
+    fn stop_monitoring(&mut self) {
+        if self.skel.maps.bss_data.as_ref().unwrap().is_monitored {
+            self.skel.maps.bss_data.as_mut().unwrap().is_monitored = false;
+        }
+    }
+
     pub fn exited(&mut self) -> bool {
         uei_exited!(&self.skel, uei)
     }
@@ -745,8 +792,13 @@ impl<'a> Scheduler<'a> {
                     let res = self.stats_req_to_res(&req)?;
                     res_ch.send(res)?;
                 }
-                Err(RecvTimeoutError::Timeout) => {}
-                Err(e) => Err(e)?,
+                Err(RecvTimeoutError::Timeout) => {
+                    self.stop_monitoring();
+                }
+                Err(e) => {
+                    self.stop_monitoring();
+                    Err(e)?
+                }
             }
             self.cleanup_introspec();
         }
@@ -811,8 +863,10 @@ fn main() -> Result<()> {
 
     init_log(&opts);
 
-    opts.proc().unwrap();
-    info!("{:#?}", opts);
+    if opts.monitor.is_none() && opts.monitor_sched_samples.is_none() {
+        opts.proc().unwrap();
+        info!("{:#?}", opts);
+    }
 
     let shutdown = Arc::new(AtomicBool::new(false));
     let shutdown_clone = shutdown.clone();
