@@ -40,6 +40,8 @@ const GSTAT_FB_CPU_USAGE: usize = bpf_intf::global_stat_id_GSTAT_FB_CPU_USAGE as
 const GSTAT_ANTISTALL: usize = bpf_intf::global_stat_id_GSTAT_ANTISTALL as usize;
 const GSTAT_SKIP_PREEMPT: usize = bpf_intf::global_stat_id_GSTAT_SKIP_PREEMPT as usize;
 const GSTAT_FIXUP_VTIME: usize = bpf_intf::global_stat_id_GSTAT_FIXUP_VTIME as usize;
+const GSTAT_PREEMPTING_MISMATCH: usize =
+    bpf_intf::global_stat_id_GSTAT_PREEMPTING_MISMATCH as usize;
 
 const LSTAT_SEL_LOCAL: usize = bpf_intf::layer_stat_id_LSTAT_SEL_LOCAL as usize;
 const LSTAT_ENQ_LOCAL: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_LOCAL as usize;
@@ -504,12 +506,18 @@ pub struct SysStats {
     pub skip_preempt: u64,
     #[stat(desc = "Number of times vtime was out of range and fixed up")]
     pub fixup_vtime: u64,
+    #[stat(desc = "Number of times cpuc->preempting_task didn't come on the CPU")]
+    pub preempting_mismatch: u64,
     #[stat(desc = "fallback CPU")]
     pub fallback_cpu: u32,
     #[stat(desc = "per-layer statistics")]
     pub fallback_cpu_util: f64,
     #[stat(desc = "fallback CPU util %")]
     pub layers: BTreeMap<String, LayerStats>,
+    #[stat(desc = "Number of gpu tasks affinitized since scheduler start")]
+    pub gpu_tasks_affinitized: u64,
+    #[stat(desc = "Time (in ms) of last affinitization run.")]
+    pub gpu_task_affinitization_ms: u64,
 }
 
 impl SysStats {
@@ -561,11 +569,14 @@ impl SysStats {
             antistall: stats.bpf_stats.gstats[GSTAT_ANTISTALL],
             skip_preempt: stats.bpf_stats.gstats[GSTAT_SKIP_PREEMPT],
             fixup_vtime: stats.bpf_stats.gstats[GSTAT_FIXUP_VTIME],
+            preempting_mismatch: stats.bpf_stats.gstats[GSTAT_PREEMPTING_MISMATCH],
             fallback_cpu: fallback_cpu as u32,
             fallback_cpu_util: stats.bpf_stats.gstats[GSTAT_FB_CPU_USAGE] as f64
                 / elapsed_ns as f64
                 * 100.0,
             layers: BTreeMap::new(),
+            gpu_tasks_affinitized: stats.gpu_tasks_affinitized,
+            gpu_task_affinitization_ms: stats.gpu_task_affinitization_ms,
         })
     }
 
@@ -602,8 +613,14 @@ impl SysStats {
 
         writeln!(
             w,
-            "skip_preempt={} antistall={} fixup_vtime={}",
-            self.skip_preempt, self.antistall, self.fixup_vtime
+            "skip_preempt={} antistall={} fixup_vtime={} preempting_mismatch={}",
+            self.skip_preempt, self.antistall, self.fixup_vtime, self.preempting_mismatch
+        )?;
+
+        writeln!(
+            w,
+            "gpu_tasks_affinitized={} gpu_task_affinitization_time={}",
+            self.gpu_tasks_affinitized, self.gpu_task_affinitization_ms
         )?;
 
         Ok(())
