@@ -26,17 +26,14 @@ enum consts {
 	MAX_CELLS = 16,
 	USAGE_HALF_LIFE = 100000000, /* 100ms */
 
-	HI_FALLBACK_DSQ = MAX_CELLS,
-	LO_FALLBACK_DSQ = MAX_CELLS + 1,
+	PCPU_BASE = 0x80000000
 };
 
 /* Statistics */
 enum cell_stat_idx {
 	CSTAT_LOCAL,
-	CSTAT_GLOBAL,
-	CSTAT_LO_FALLBACK_Q,
-	CSTAT_HI_FALLBACK_Q,
-	CSTAT_DEFAULT_Q,
+	CSTAT_CPU_DSQ,
+	CSTAT_CELL_DSQ,
 	CSTAT_AFFN_VIOL,
 	NR_CSTATS,
 };
@@ -45,6 +42,7 @@ struct cpu_ctx {
 	u64 cstats[MAX_CELLS][NR_CSTATS];
 	u64 cell_cycles[MAX_CELLS];
 	u32 cell;
+	u64 vtime_now;
 };
 
 struct cgrp_ctx {
@@ -58,8 +56,6 @@ struct cgrp_ctx {
 struct cell {
 	// current vtime of the cell
 	u64 vtime_now;
-	// which dsq the cell uses
-	u32 dsq;
 	// Whether or not the cell is used or not
 	u32 in_use;
 };
