@@ -11,6 +11,7 @@ use scx_utils::Topology;
 pub use scx_utils::NR_CPU_IDS;
 
 use clap::Parser;
+use clap::ValueEnum;
 
 lazy_static::lazy_static! {
         pub static ref TOPO: Topology = Topology::new().unwrap();
@@ -26,6 +27,23 @@ fn get_default_llc_runs() -> u64 {
     llc_runs as u64
 }
 
+#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, ValueEnum)]
+pub enum LbMode {
+    /// load of the LLC
+    Load,
+    /// number of tasks queued
+    NrQueued,
+}
+
+impl LbMode {
+    pub fn as_i32(&self) -> i32 {
+        match self {
+            LbMode::Load => bpf_intf::p2dq_lb_mode_PICK2_LOAD as i32,
+            LbMode::NrQueued => bpf_intf::p2dq_lb_mode_PICK2_NR_QUEUED as i32,
+        }
+    }
+}
+
 #[derive(Debug, Parser)]
 pub struct SchedulerOpts {
     /// Disables per-cpu kthreads directly dispatched into local dsqs.
@@ -44,7 +62,7 @@ pub struct SchedulerOpts {
     #[clap(long, action = clap::ArgAction::SetTrue)]
     pub deadline: bool,
 
-    /// *DEPRECATED* Disables eager pick2 load balancing.
+    /// ***DEPRECATED*** Disables eager pick2 load balancing.
     #[clap(short = 'e', long, help="DEPRECATED", action = clap::ArgAction::SetTrue)]
     pub eager_load_balance: bool,
 
@@ -61,7 +79,7 @@ pub struct SchedulerOpts {
     #[clap(short = 'y', long, action = clap::ArgAction::SetTrue)]
     pub interactive_sticky: bool,
 
-    /// Interactive tasks are FIFO scheduled
+    /// ***DEPRECATED*** Interactive tasks are FIFO scheduled
     #[clap(long, action = clap::ArgAction::SetTrue)]
     pub interactive_fifo: bool,
 
@@ -82,7 +100,16 @@ pub struct SchedulerOpts {
     #[clap(long, action = clap::ArgAction::SetTrue)]
     pub keep_running: bool,
 
-    /// Use a separate DSQ for interactive tasks
+    /// Use a arena based queues (ATQ) for task queueing.
+    #[clap(long, default_value_t = false, action = clap::ArgAction::Set)]
+    pub atq_enabled: bool,
+
+    /// Schedule based on preferred core values available on some x86 systems with the appropriate
+    /// CPU frequency governor (ex: amd-pstate).
+    #[clap(long, default_value_t = false, action = clap::ArgAction::Set)]
+    pub cpu_priority: bool,
+
+    /// ***DEPRECATED*** Use a separate DSQ for interactive tasks
     #[clap(long, default_value_t = true, action = clap::ArgAction::Set)]
     pub interactive_dsq: bool,
 
@@ -98,6 +125,10 @@ pub struct SchedulerOpts {
     #[clap(long, action = clap::ArgAction::SetTrue)]
     pub select_idle_in_enqueue: bool,
 
+    /// Allow queued wakeup.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    pub queued_wakeup: bool,
+
     /// Set idle QoS resume latency based in microseconds.
     #[clap(long)]
     pub idle_resume_us: Option<u32>,
@@ -110,16 +141,25 @@ pub struct SchedulerOpts {
     #[clap(short = 's', long, default_value = "100")]
     pub min_slice_us: u64,
 
+    /// ***DEPRECATED*** Load balance mode
+    #[arg(value_enum, long, default_value_t = LbMode::Load)]
+    pub lb_mode: LbMode,
+
     /// Slack factor for load balancing, load balancing is not performed if load is within slack
     /// factor percent.
     #[clap(long, default_value = "5", value_parser = clap::value_parser!(u64).range(0..99))]
     pub lb_slack_factor: u64,
 
-    /// Number of runs on the LLC before a task becomes eligbile for pick2 migration on the wakeup
+    /// ***DEPRECATED** Number of runs on the LLC before a task becomes eligbile for pick2 migration on the wakeup
     /// path.
     #[clap(short = 'l', long, default_value_t = get_default_llc_runs())]
     pub min_llc_runs_pick2: u64,
 
+    /// Saturated percent is the percent at which the system is considered saturated in terms of
+    /// free CPUs.
+    #[clap(long, default_value_t = 5)]
+    pub saturated_percent: u32,
+
     /// Manual definition of slice intervals in microseconds for DSQs, must be equal to number of
     /// dumb_queues.
     #[clap(short = 't', long, value_parser = clap::value_parser!(u64), default_values_t = [0;0])]
@@ -131,7 +171,7 @@ pub struct SchedulerOpts {
 
     /// Minimum number of queued tasks to use pick2 balancing, 0 to always enabled.
     #[clap(short = 'm', long, default_value = "0")]
-    pub min_nr_queued_pick2: u32,
+    pub min_nr_queued_pick2: u64,
 
     /// Number of dumb DSQs.
     #[clap(short = 'q', long, default_value = "3")]
@@ -155,6 +195,7 @@ pub fn dsq_slice_ns(dsq_index: u64, min_slice_us: u64, dsq_shift: u64) -> u64 {
 macro_rules! init_open_skel {
     ($skel: expr, $opts: expr, $verbose: expr) => {
         'block: {
+            let skel = $skel;
             let opts: &$crate::SchedulerOpts = $opts;
             let verbose: u8 = $verbose;
 
@@ -181,14 +222,14 @@ macro_rules! init_open_skel {
                 }
                 for (i, slice) in opts.dsq_time_slices.iter().enumerate() {
                     ::log::info!("DSQ[{}] slice_ns {}", i, slice * 1000);
-                    $skel.maps.bss_data.dsq_time_slices[i] = slice * 1000;
+                    skel.maps.bss_data.as_mut().unwrap().dsq_time_slices[i] = slice * 1000;
                 }
             } else {
                 for i in 0..=opts.dumb_queues - 1 {
                     let slice_ns =
                         $crate::dsq_slice_ns(i as u64, opts.min_slice_us, opts.dsq_shift);
                     ::log::info!("DSQ[{}] slice_ns {}", i, slice_ns);
-                    $skel.maps.bss_data.dsq_time_slices[i] = slice_ns;
+                    skel.maps.bss_data.as_mut().unwrap().dsq_time_slices[i] = slice_ns;
                 }
             }
             if opts.autoslice {
@@ -200,40 +241,53 @@ macro_rules! init_open_skel {
                 }
             }
 
-            $skel.maps.rodata_data.interactive_ratio = opts.interactive_ratio as u32;
-            $skel.maps.rodata_data.min_slice_us = opts.min_slice_us;
-            $skel.maps.rodata_data.min_nr_queued_pick2 = opts.min_nr_queued_pick2;
-            $skel.maps.rodata_data.min_llc_runs_pick2 = opts.min_llc_runs_pick2;
-            $skel.maps.rodata_data.dsq_shift = opts.dsq_shift as u64;
-            $skel.maps.rodata_data.kthreads_local = !opts.disable_kthreads_local;
-            $skel.maps.rodata_data.nr_cpus = *$crate::NR_CPU_IDS as u32;
-            $skel.maps.rodata_data.nr_dsqs_per_llc = opts.dumb_queues as u32;
-            $skel.maps.rodata_data.nr_cpu_ids = *NR_CPU_IDS as u32;
-            $skel.maps.rodata_data.init_dsq_index = opts.init_dsq_index as i32;
-            $skel.maps.rodata_data.nr_llcs = $crate::TOPO.all_llcs.clone().keys().len() as u32;
-            $skel.maps.rodata_data.nr_nodes = $crate::TOPO.nodes.clone().keys().len() as u32;
-            $skel.maps.rodata_data.lb_slack_factor = opts.lb_slack_factor;
+            // topo config
+            let rodata = skel.maps.rodata_data.as_mut().unwrap();
+            rodata.topo_config.nr_cpus = *$crate::NR_CPU_IDS as u32;
+            rodata.topo_config.nr_llcs = $crate::TOPO.all_llcs.clone().keys().len() as u32;
+            rodata.topo_config.nr_nodes = $crate::TOPO.nodes.clone().keys().len() as u32;
+            rodata.topo_config.smt_enabled = MaybeUninit::new($crate::TOPO.smt_enabled);
+            rodata.topo_config.has_little_cores = MaybeUninit::new($crate::TOPO.has_little_cores());
 
-            $skel.maps.rodata_data.autoslice = opts.autoslice;
-            $skel.maps.rodata_data.debug = verbose as u32;
-            $skel.maps.rodata_data.deadline_scheduling = opts.deadline;
-            $skel.maps.rodata_data.dispatch_pick2_disable = opts.dispatch_pick2_disable;
-            $skel.maps.rodata_data.dispatch_lb_busy = opts.dispatch_lb_busy;
-            $skel.maps.rodata_data.dispatch_lb_interactive = opts.dispatch_lb_interactive;
-            $skel.maps.rodata_data.eager_load_balance = !opts.eager_load_balance;
-            $skel.maps.rodata_data.freq_control = opts.freq_control;
-            $skel.maps.rodata_data.has_little_cores = $crate::TOPO.has_little_cores();
-            $skel.maps.rodata_data.interactive_dsq = opts.interactive_dsq;
-            $skel.maps.rodata_data.interactive_sticky = opts.interactive_sticky;
-            $skel.maps.rodata_data.interactive_fifo = opts.interactive_fifo;
-            $skel.maps.rodata_data.keep_running_enabled = opts.keep_running;
-            $skel.maps.rodata_data.max_dsq_pick2 = opts.max_dsq_pick2;
-            $skel.maps.rodata_data.smt_enabled = $crate::TOPO.smt_enabled;
-            $skel.maps.rodata_data.select_idle_in_enqueue = opts.select_idle_in_enqueue;
-            $skel.maps.rodata_data.wakeup_lb_busy = opts.wakeup_lb_busy;
-            $skel.maps.rodata_data.wakeup_llc_migrations = opts.wakeup_llc_migrations;
-            $skel.maps.rodata_data.max_exec_ns =
-                2 * $skel.maps.bss_data.dsq_time_slices[opts.dumb_queues - 1];
+            // timeline config
+            rodata.timeline_config.min_slice_us = opts.min_slice_us;
+            rodata.timeline_config.max_exec_ns =
+                2 * skel.maps.bss_data.as_ref().unwrap().dsq_time_slices[opts.dumb_queues - 1];
+            rodata.timeline_config.autoslice = MaybeUninit::new(opts.autoslice);
+            rodata.timeline_config.deadline = MaybeUninit::new(opts.deadline);
+
+            // load balance config
+            rodata.lb_config.slack_factor = opts.lb_slack_factor;
+            rodata.lb_config.min_nr_queued_pick2 = opts.min_nr_queued_pick2;
+            rodata.lb_config.max_dsq_pick2 = MaybeUninit::new(opts.max_dsq_pick2);
+            rodata.lb_config.eager_load_balance = MaybeUninit::new(!opts.eager_load_balance);
+            rodata.lb_config.dispatch_pick2_disable = MaybeUninit::new(opts.dispatch_pick2_disable);
+            rodata.lb_config.dispatch_lb_busy = opts.dispatch_lb_busy;
+            rodata.lb_config.dispatch_lb_interactive =
+                MaybeUninit::new(opts.dispatch_lb_interactive);
+            rodata.lb_config.wakeup_lb_busy = opts.wakeup_lb_busy;
+            rodata.lb_config.wakeup_llc_migrations = MaybeUninit::new(opts.wakeup_llc_migrations);
+
+            // p2dq config
+            rodata.p2dq_config.interactive_ratio = opts.interactive_ratio as u32;
+            rodata.p2dq_config.dsq_shift = opts.dsq_shift as u64;
+            rodata.p2dq_config.kthreads_local = MaybeUninit::new(!opts.disable_kthreads_local);
+            rodata.p2dq_config.nr_dsqs_per_llc = opts.dumb_queues as u32;
+            rodata.p2dq_config.init_dsq_index = opts.init_dsq_index as i32;
+            rodata.p2dq_config.saturated_percent = opts.saturated_percent;
+
+            rodata.p2dq_config.atq_enabled = MaybeUninit::new(
+                opts.atq_enabled && compat::ksym_exists("bpf_spin_unlock").unwrap_or(false),
+            );
+            rodata.p2dq_config.cpu_priority = MaybeUninit::new(opts.cpu_priority);
+            rodata.p2dq_config.freq_control = MaybeUninit::new(opts.freq_control);
+            rodata.p2dq_config.interactive_sticky = MaybeUninit::new(opts.interactive_sticky);
+            rodata.p2dq_config.keep_running_enabled = MaybeUninit::new(opts.keep_running);
+            rodata.p2dq_config.select_idle_in_enqueue =
+                MaybeUninit::new(opts.select_idle_in_enqueue);
+
+            rodata.debug = verbose as u32;
+            rodata.nr_cpu_ids = *NR_CPU_IDS as u32;
 
             Ok(())
         }
@@ -244,17 +298,17 @@ macro_rules! init_open_skel {
 macro_rules! init_skel {
     ($skel: expr) => {
         for cpu in $crate::TOPO.all_cpus.values() {
-            $skel.maps.bss_data.big_core_ids[cpu.id] =
+            $skel.maps.bss_data.as_mut().unwrap().big_core_ids[cpu.id] =
                 if cpu.core_type == ($crate::CoreType::Big { turbo: true }) {
                     1
                 } else {
                     0
                 };
-            $skel.maps.bss_data.cpu_llc_ids[cpu.id] = cpu.llc_id as u64;
-            $skel.maps.bss_data.cpu_node_ids[cpu.id] = cpu.node_id as u64;
+            $skel.maps.bss_data.as_mut().unwrap().cpu_llc_ids[cpu.id] = cpu.llc_id as u64;
+            $skel.maps.bss_data.as_mut().unwrap().cpu_node_ids[cpu.id] = cpu.node_id as u64;
         }
         for llc in $crate::TOPO.all_llcs.values() {
-            $skel.maps.bss_data.llc_ids[llc.id] = llc.id as u64;
+            $skel.maps.bss_data.as_mut().unwrap().llc_ids[llc.id] = llc.id as u64;
         }
     };
 }
