@@ -38,10 +38,13 @@ use scx_utils::{Core, Llc};
 use std::ffi::c_ulong;
 
 use bpf_intf::stat_idx_P2DQ_NR_STATS;
+use bpf_intf::stat_idx_P2DQ_STAT_ATQ_ENQ;
+use bpf_intf::stat_idx_P2DQ_STAT_ATQ_REENQ;
 use bpf_intf::stat_idx_P2DQ_STAT_DIRECT;
 use bpf_intf::stat_idx_P2DQ_STAT_DISPATCH_PICK2;
 use bpf_intf::stat_idx_P2DQ_STAT_DSQ_CHANGE;
 use bpf_intf::stat_idx_P2DQ_STAT_DSQ_SAME;
+use bpf_intf::stat_idx_P2DQ_STAT_ENQ_CPU;
 use bpf_intf::stat_idx_P2DQ_STAT_ENQ_INTR;
 use bpf_intf::stat_idx_P2DQ_STAT_ENQ_LLC;
 use bpf_intf::stat_idx_P2DQ_STAT_ENQ_MIG;
@@ -112,10 +115,10 @@ impl<'a> Scheduler<'a> {
         let mut open_skel = scx_ops_open!(skel_builder, open_object, p2dq).unwrap();
         scx_p2dq::init_open_skel!(&mut open_skel, opts, verbose)?;
 
-        match *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP {
-            0 => info!("Kernel does not support queued wakeup optimization."),
-            v => open_skel.struct_ops.p2dq_mut().flags |= v,
-        };
+        if opts.queued_wakeup {
+            open_skel.struct_ops.p2dq_mut().flags |= *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP;
+        }
+        open_skel.struct_ops.p2dq_mut().flags |= *compat::SCX_OPS_KEEP_BUILTIN_IDLE;
 
         let mut skel = scx_ops_load!(open_skel, p2dq, uei)?;
 
@@ -146,14 +149,17 @@ impl<'a> Scheduler<'a> {
             stats[stat as usize] = sum;
         }
         Metrics {
+            atq_enq: stats[stat_idx_P2DQ_STAT_ATQ_ENQ as usize],
+            atq_reenq: stats[stat_idx_P2DQ_STAT_ATQ_REENQ as usize],
             direct: stats[stat_idx_P2DQ_STAT_DIRECT as usize],
             idle: stats[stat_idx_P2DQ_STAT_IDLE as usize],
-            sched_mode: self.skel.maps.bss_data.sched_mode,
+            sched_mode: self.skel.maps.bss_data.as_ref().unwrap().sched_mode,
             dsq_change: stats[stat_idx_P2DQ_STAT_DSQ_CHANGE as usize],
             same_dsq: stats[stat_idx_P2DQ_STAT_DSQ_SAME as usize],
             keep: stats[stat_idx_P2DQ_STAT_KEEP as usize],
-            enq_llc: stats[stat_idx_P2DQ_STAT_ENQ_LLC as usize],
+            enq_cpu: stats[stat_idx_P2DQ_STAT_ENQ_CPU as usize],
             enq_intr: stats[stat_idx_P2DQ_STAT_ENQ_INTR as usize],
+            enq_llc: stats[stat_idx_P2DQ_STAT_ENQ_LLC as usize],
             enq_mig: stats[stat_idx_P2DQ_STAT_ENQ_MIG as usize],
             select_pick2: stats[stat_idx_P2DQ_STAT_SELECT_PICK2 as usize],
             dispatch_pick2: stats[stat_idx_P2DQ_STAT_DISPATCH_PICK2 as usize],
