@@ -17,12 +17,18 @@ use serde::Serialize;
 pub struct Metrics {
     #[stat(desc = "Scheduler mode")]
     pub sched_mode: u32,
+    #[stat(desc = "Number of times a task was enqueued to a ATQ")]
+    pub atq_enq: u64,
+    #[stat(desc = "Number of times a task was re-enqueued to a ATQ")]
+    pub atq_reenq: u64,
     #[stat(desc = "Number of times tasks have switched DSQs")]
     pub dsq_change: u64,
     #[stat(desc = "Number of times tasks have stayed on the same DSQ")]
     pub same_dsq: u64,
     #[stat(desc = "Number of times a task kept running")]
     pub keep: u64,
+    #[stat(desc = "Number of times a task was enqueued to CPUC DSQ")]
+    pub enq_cpu: u64,
     #[stat(desc = "Number of times a task was enqueued to LLC DSQ")]
     pub enq_llc: u64,
     #[stat(desc = "Number of times a task was enqueued to interactive DSQ")]
@@ -53,12 +59,15 @@ impl Metrics {
     fn format<W: Write>(&self, w: &mut W) -> Result<()> {
         writeln!(
             w,
-            "direct/idle/keep {}/{}/{}\n\tdsq same/migrate {}/{}\n\tenq llc/intr/mig {}/{}/{}",
+            "direct/idle/keep {}/{}/{}\n\tdsq same/migrate {}/{}\n\tatq enq/reenq {}/{}\n\tenq cpu/llc/intr/mig {}/{}/{}/{}",
             self.direct,
             self.idle,
             self.keep,
             self.same_dsq,
             self.dsq_change,
+            self.atq_enq,
+            self.atq_reenq,
+            self.enq_cpu,
             self.enq_llc,
             self.enq_intr,
             self.enq_mig,
@@ -79,11 +88,14 @@ impl Metrics {
 
     fn delta(&self, rhs: &Self) -> Self {
         Self {
+            atq_enq: self.atq_enq - rhs.atq_enq,
+            atq_reenq: self.atq_reenq - rhs.atq_reenq,
             direct: self.direct - rhs.direct,
             idle: self.idle - rhs.idle,
             dsq_change: self.dsq_change - rhs.dsq_change,
             same_dsq: self.same_dsq - rhs.same_dsq,
             keep: self.keep - rhs.keep,
+            enq_cpu: self.enq_cpu - rhs.enq_cpu,
             enq_llc: self.enq_llc - rhs.enq_llc,
             enq_intr: self.enq_intr - rhs.enq_intr,
             enq_mig: self.enq_mig - rhs.enq_mig,
