@@ -8,84 +8,111 @@ pub mod bpf_intf;
 pub mod bpf_skel;
 mod bpf_stats;
 pub mod cli;
+mod columns;
 pub mod config;
 mod cpu_data;
-pub mod cpu_stats;
+mod cpu_stats;
 pub mod edm;
 mod event_data;
 mod keymap;
+pub mod layered_util;
 mod llc_data;
 pub mod mangoapp;
+mod mem_stats;
+mod network_stats;
 mod node_data;
 mod perfetto_trace;
+mod power_data;
+mod proc_data;
 pub mod profiling_events;
-mod search;
+pub mod search;
 mod stats;
+mod symbol_data;
 mod theme;
+mod thread_data;
 pub mod tracer;
 mod tui;
-mod util;
+pub mod util;
 
 pub use crate::bpf_skel::types::bpf_event;
 pub use app::App;
 pub use bpf_skel::*;
+pub use columns::Columns;
 pub use cpu_data::CpuData;
-pub use cpu_stats::CpuStatTracker;
+pub use cpu_stats::{CpuStatSnapshot, CpuStatTracker};
 pub use event_data::EventData;
 pub use keymap::Key;
 pub use keymap::KeyMap;
 pub use llc_data::LlcData;
+pub use mem_stats::MemStatSnapshot;
+pub use network_stats::NetworkStatSnapshot;
 pub use node_data::NodeData;
 pub use perfetto_trace::PerfettoTraceManager;
+pub use power_data::{
+    CStateInfo, CorePowerData, PowerDataCollector, PowerSnapshot, SystemPowerData,
+};
+pub use proc_data::ProcData;
 pub use profiling_events::{
     available_kprobe_events, available_perf_events, get_default_events, KprobeEvent, PerfEvent,
     ProfilingEvent,
 };
-pub use search::Search;
 pub use stats::StatAggregation;
 pub use stats::VecStats;
 pub use theme::AppTheme;
+pub use thread_data::ThreadData;
 pub use tui::Event;
 pub use tui::Tui;
-pub use util::format_hz;
-pub use util::read_file_string;
-pub use util::sanitize_nbsp;
 
 pub use plain::Plain;
 // Generate serialization types for handling events from the bpf ring buffer.
 unsafe impl Plain for crate::bpf_skel::types::bpf_event {}
 
 use smartstring::alias::String as SsoString;
+use std::collections::BTreeMap;
 
 pub const APP: &str = "scxtop";
 pub const TRACE_FILE_PREFIX: &str = "scxtop_trace";
 pub const STATS_SOCKET_PATH: &str = "/var/run/scx/root/stats";
-pub const LICENSE: &str = "Copyright (c) Meta Platforms, Inc. and affiliates.
-
-This software may be used and distributed according to the terms of the 
-GNU General Public License version 2.";
+pub const LICENSE: &str = concat!(
+    "Copyright (c) Meta Platforms, Inc. and affiliates.\n",
+    "\n",
+    "This software may be used and distributed according to the terms of the \n",
+    "GNU General Public License version 2."
+);
 pub const SCHED_NAME_PATH: &str = "/sys/kernel/sched_ext/root/ops";
 
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub enum AppState {
     /// Application is in the default state.
     Default,
-    /// Application is in the PerfEvent list state.
-    PerfEvent,
-    /// Application is in the KprobeEvent list state.
-    KprobeEvent,
     /// Application is in the help state.
     Help,
+    /// Application is in the KprobeEvent list state.
+    KprobeEvent,
     /// Application is in the Llc state.
     Llc,
+    /// Application is in the mangoapp state.
+    MangoApp,
+    /// Application is in the Memory state.
+    Memory,
+    /// Application is in the network state.
+    Network,
     /// Application is in the NUMA node state.
     Node,
+    /// Application is in the paused state.
+    Pause,
+    /// Application is in the PerfEvent list state.
+    PerfEvent,
+    /// Application is in the perf top view state.
+    PerfTop,
+    /// Application is in the Power state.
+    Power,
+    /// Application is in the Process state.
+    Process,
     /// Application is in the scheduler state.
     Scheduler,
     /// Application is in the tracing  state.
     Tracing,
-    /// Application is in the mangoapp state.
-    MangoApp,
 }
 
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
@@ -113,15 +140,67 @@ impl std::fmt::Display for ViewState {
     }
 }
 
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub enum ComponentViewState {
+    /// Component view is hidden in the default view
+    Hidden,
+    /// Component view is shown in the default view (current behavior)
+    Default,
+    /// Switch to detailed component view
+    Detail,
+}
+
+impl ComponentViewState {
+    /// Returns the next ComponentViewState, cycling through the values.
+    pub fn next(&self) -> Self {
+        match self {
+            ComponentViewState::Hidden => ComponentViewState::Default,
+            ComponentViewState::Default => ComponentViewState::Detail,
+            ComponentViewState::Detail => ComponentViewState::Hidden,
+        }
+    }
+}
+
+impl std::fmt::Display for ComponentViewState {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self {
+            ComponentViewState::Hidden => write!(f, "hidden"),
+            ComponentViewState::Default => write!(f, "default"),
+            ComponentViewState::Detail => write!(f, "detail"),
+        }
+    }
+}
+
+#[derive(Debug, Clone, PartialEq)]
+pub enum FilterItem {
+    String(String),
+    Int(i32),
+}
+impl FilterItem {
+    pub fn as_string(&self) -> String {
+        match self {
+            FilterItem::String(s) => s.clone(),
+            FilterItem::Int(int) => int.to_string(),
+        }
+    }
+
+    pub fn as_int(&self) -> Option<i32> {
+        match self {
+            FilterItem::Int(int) => Some(*int),
+            FilterItem::String(_) => None,
+        }
+    }
+}
+
 #[derive(Default, Debug, Clone)]
-pub struct FilteredEventState {
-    pub list: Vec<String>,
+pub struct FilteredState {
+    pub list: Vec<FilterItem>,
     pub count: u16,
     pub scroll: u16,
     pub selected: usize,
 }
 
-impl FilteredEventState {
+impl FilteredState {
     pub fn reset(&mut self) {
         self.list.clear();
         self.count = 0;
@@ -151,9 +230,13 @@ pub struct ForkAction {
     pub ts: u64,
     pub cpu: u32,
     pub parent_pid: u32,
+    pub parent_tgid: u32,
     pub child_pid: u32,
+    pub child_tgid: u32,
     pub parent_comm: SsoString,
     pub child_comm: SsoString,
+    pub parent_layer_id: i32,
+    pub child_layer_id: i32,
 }
 
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
@@ -162,6 +245,16 @@ pub struct ExecAction {
     pub cpu: u32,
     pub old_pid: u32,
     pub pid: u32,
+    pub layer_id: i32,
+}
+
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct WaitAction {
+    pub ts: u64,
+    pub cpu: u32,
+    pub comm: SsoString,
+    pub pid: u32,
+    pub prio: i32,
 }
 
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
@@ -177,6 +270,7 @@ pub struct SchedSwitchAction {
     pub next_pid: u32,
     pub next_tgid: u32,
     pub next_prio: i32,
+    pub next_layer_id: i32,
     pub next_comm: SsoString,
     pub prev_dsq_id: u64,
     pub prev_used_slice_ns: u64,
@@ -186,6 +280,7 @@ pub struct SchedSwitchAction {
     pub prev_prio: i32,
     pub prev_comm: SsoString,
     pub prev_state: u64,
+    pub prev_layer_id: i32,
 }
 
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
@@ -212,6 +307,14 @@ pub struct SchedMigrateTaskAction {
     pub comm: SsoString,
 }
 
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct SchedHangAction {
+    pub ts: u64,
+    pub cpu: u32,
+    pub comm: SsoString,
+    pub pid: u32,
+}
+
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct SoftIRQAction {
     pub cpu: u32,
@@ -275,12 +378,6 @@ pub struct HwPressureAction {
     pub cpu: u32,
 }
 
-#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
-pub struct PstateSampleAction {
-    pub cpu: u32,
-    pub busy: u32,
-}
-
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct KprobeAction {
     pub ts: u64,
@@ -289,6 +386,21 @@ pub struct KprobeAction {
     pub instruction_pointer: u64,
 }
 
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct SystemStatAction {
+    pub ts: u64,
+    pub cpu_data_prev: BTreeMap<usize, CpuStatSnapshot>,
+    pub cpu_data_current: BTreeMap<usize, CpuStatSnapshot>,
+    pub mem_info: MemStatSnapshot,
+}
+
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct UpdateColVisibilityAction {
+    pub table: String,
+    pub col: String,
+    pub visible: bool,
+}
+
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct MangoAppAction {
     pub pid: u32,
@@ -302,6 +414,20 @@ pub struct MangoAppAction {
     pub display_refresh: u16,
 }
 
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct PerfSampleAction {
+    pub ts: u64,
+    pub cpu: u32,
+    pub pid: u32,
+    pub instruction_pointer: u64,
+    pub cpu_id: u32,
+    pub is_kernel: bool,
+    pub kernel_stack: Vec<u64>,
+    pub user_stack: Vec<u64>,
+    pub layer_id: i32,
+}
+
+#[allow(clippy::large_enum_variant)]
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub enum Action {
     Backspace,
@@ -309,6 +435,7 @@ pub enum Action {
     ClearEvent,
     CpuhpEnter(CpuhpEnterAction),
     CpuhpExit(CpuhpExitAction),
+    SystemStat(SystemStatAction),
     DecBpfSampleRate,
     DecTickRate,
     Down,
@@ -317,6 +444,7 @@ pub enum Action {
     Esc,
     Exec(ExecAction),
     Exit(ExitAction),
+    Filter,
     Fork(ForkAction),
     Kprobe(KprobeAction),
     GpuMem(GpuMemAction),
@@ -331,8 +459,10 @@ pub enum Action {
     NextViewState,
     PageDown,
     PageUp,
+    PerfSample(PerfSampleAction),
+    PerfSampleRateIncrease,
+    PerfSampleRateDecrease,
     PrevEvent,
-    PstateSample(PstateSampleAction),
     Quit,
     RequestTrace,
     TraceStarted(TraceStartedAction),
@@ -340,6 +470,7 @@ pub enum Action {
     ReloadStatsClient,
     SaveConfig,
     SchedCpuPerfSet(SchedCpuPerfSetAction),
+    SchedHang(SchedHangAction),
     SchedMigrateTask(SchedMigrateTaskAction),
     SchedReg,
     SchedStats(String),
@@ -357,6 +488,8 @@ pub enum Action {
     ToggleHwPressure,
     ToggleUncoreFreq,
     Up,
+    UpdateColVisibility(UpdateColVisibilityAction),
+    Wait(WaitAction),
     None,
 }
 
@@ -478,12 +611,30 @@ impl TryFrom<&bpf_event> for Action {
                     comm: comm.into(),
                 }))
             }
-            bpf_intf::event_type_EXEC => Ok(Action::Exec(ExecAction {
-                ts: event.ts,
-                cpu: event.cpu,
-                old_pid: unsafe { event.event.exec.old_pid },
-                pid: unsafe { event.event.exec.pid },
-            })),
+            #[allow(non_upper_case_globals)]
+            bpf_intf::event_type_SCHED_HANG => {
+                let hang = unsafe { &event.event.hang };
+
+                let comm = String::from_utf8_lossy(&hang.comm);
+
+                Ok(Action::SchedHang(SchedHangAction {
+                    ts: event.ts,
+                    cpu: event.cpu,
+                    comm: comm.into(),
+                    pid: hang.pid,
+                }))
+            }
+            bpf_intf::event_type_EXEC => {
+                let exec = unsafe { &event.event.exec };
+
+                Ok(Action::Exec(ExecAction {
+                    ts: event.ts,
+                    cpu: event.cpu,
+                    old_pid: exec.old_pid,
+                    pid: exec.pid,
+                    layer_id: exec.layer_id,
+                }))
+            }
             bpf_intf::event_type_EXIT => {
                 let exit = unsafe { &event.event.exit };
                 let comm = String::from_utf8_lossy(&exit.comm);
@@ -507,9 +658,13 @@ impl TryFrom<&bpf_event> for Action {
                     ts: event.ts,
                     cpu: event.cpu,
                     parent_pid: fork.parent_pid,
+                    parent_tgid: fork.parent_tgid,
                     child_pid: fork.child_pid,
+                    child_tgid: fork.child_tgid,
                     parent_comm: parent_comm.into(),
                     child_comm: child_comm.into(),
+                    parent_layer_id: fork.parent_layer_id,
+                    child_layer_id: fork.child_layer_id,
                 }))
             }
             #[allow(non_upper_case_globals)]
@@ -524,10 +679,35 @@ impl TryFrom<&bpf_event> for Action {
                 }))
             }
             #[allow(non_upper_case_globals)]
-            bpf_intf::event_type_PSTATE_SAMPLE => Ok(Action::PstateSample(PstateSampleAction {
-                cpu: event.cpu,
-                busy: unsafe { event.event.pstate.busy },
-            })),
+            bpf_intf::event_type_PERF_SAMPLE => {
+                let perf_sample = unsafe { &event.event.perf_sample };
+
+                // Extract kernel stack trace
+                let kernel_stack: Vec<u64> = if perf_sample.kernel_stack_size > 0 {
+                    perf_sample.kernel_stack[0..perf_sample.kernel_stack_size as usize].to_vec()
+                } else {
+                    Vec::new()
+                };
+
+                // Extract user stack trace
+                let user_stack: Vec<u64> = if perf_sample.user_stack_size > 0 {
+                    perf_sample.user_stack[0..perf_sample.user_stack_size as usize].to_vec()
+                } else {
+                    Vec::new()
+                };
+
+                Ok(Action::PerfSample(PerfSampleAction {
+                    ts: event.ts,
+                    cpu: event.cpu,
+                    pid: perf_sample.pid,
+                    instruction_pointer: perf_sample.instruction_pointer,
+                    cpu_id: perf_sample.cpu_id,
+                    is_kernel: unsafe { perf_sample.is_kernel.assume_init() },
+                    kernel_stack,
+                    user_stack,
+                    layer_id: perf_sample.layer_id,
+                }))
+            }
             #[allow(non_upper_case_globals)]
             bpf_intf::event_type_SCHED_SWITCH => {
                 let sched_switch = unsafe { &event.event.sched_switch };
@@ -546,6 +726,7 @@ impl TryFrom<&bpf_event> for Action {
                     next_pid: sched_switch.next_pid,
                     next_tgid: sched_switch.next_tgid,
                     next_prio: sched_switch.next_prio,
+                    next_layer_id: sched_switch.next_layer_id,
                     next_comm: next_comm.into(),
                     prev_dsq_id: sched_switch.prev_dsq_id,
                     prev_used_slice_ns: sched_switch.prev_used_slice_ns,
@@ -555,6 +736,7 @@ impl TryFrom<&bpf_event> for Action {
                     prev_comm: prev_comm.into(),
                     prev_prio: sched_switch.prev_prio,
                     prev_state: sched_switch.prev_state,
+                    prev_layer_id: sched_switch.prev_layer_id,
                 }))
             }
             #[allow(non_upper_case_globals)]
@@ -571,6 +753,18 @@ impl TryFrom<&bpf_event> for Action {
                 let action = Action::TraceStopped(TraceStoppedAction { ts: event.ts });
                 Ok(action)
             }
+            bpf_intf::event_type_WAIT => {
+                let wait = unsafe { &event.event.wait };
+                let comm = String::from_utf8_lossy(&wait.comm);
+
+                Ok(Action::Wait(WaitAction {
+                    ts: event.ts,
+                    cpu: event.cpu,
+                    comm: comm.into(),
+                    pid: wait.pid,
+                    prio: wait.prio,
+                }))
+            }
             _ => Err(()),
         }
     }
@@ -580,24 +774,30 @@ impl std::fmt::Display for Action {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
             Action::SetState(AppState::Default) => write!(f, "AppStateDefault"),
+            Action::SetState(AppState::Pause) => write!(f, "AppStatePause"),
             Action::SetState(AppState::PerfEvent) => write!(f, "AppStatePerfEvent"),
+            Action::SetState(AppState::Process) => write!(f, "AppStateProcess"),
             Action::SetState(AppState::KprobeEvent) => write!(f, "AppStateKprobeEvent"),
             Action::SetState(AppState::MangoApp) => write!(f, "AppStateMangoApp"),
+            Action::Filter => write!(f, "Filter"),
+            Action::UpdateColVisibility(_) => write!(f, "UpdateColVisibility"),
             Action::ToggleCpuFreq => write!(f, "ToggleCpuFreq"),
             Action::ToggleUncoreFreq => write!(f, "ToggleUncoreFreq"),
             Action::ToggleLocalization => write!(f, "ToggleLocalization"),
             Action::ToggleHwPressure => write!(f, "ToggleHwPressure"),
             Action::SetState(AppState::Help) => write!(f, "AppStateHelp"),
             Action::SetState(AppState::Llc) => write!(f, "AppStateLlc"),
+            Action::SetState(AppState::Network) => write!(f, "AppStateNetwork"),
             Action::SetState(AppState::Node) => write!(f, "AppStateNode"),
             Action::SetState(AppState::Scheduler) => write!(f, "AppStateScheduler"),
             Action::SaveConfig => write!(f, "SaveConfig"),
             Action::RequestTrace => write!(f, "RequestTrace"),
             Action::TraceStarted(_) => write!(f, "TraceStarted"),
-            Action::TraceStopped(_) => write!(f, "TraceStopped"),
-            Action::ClearEvent => write!(f, "ClearEvent"),
+            Action::PerfSampleRateIncrease => write!(f, "PerfSampleRateIncrease"),
+            Action::PerfSampleRateDecrease => write!(f, "PerfSampleRateDecrease"),
+            Action::PageUp => write!(f, "PageUp"),
+            Action::PageDown => write!(f, "PageDown"),
             Action::PrevEvent => write!(f, "PrevEvent"),
-            Action::NextEvent => write!(f, "NextEvent"),
             Action::Quit => write!(f, "Quit"),
             Action::ChangeTheme => write!(f, "ChangeTheme"),
             Action::DecTickRate => write!(f, "DecTickRate"),
@@ -607,10 +807,8 @@ impl std::fmt::Display for Action {
             Action::NextViewState => write!(f, "NextViewState"),
             Action::Down => write!(f, "Down"),
             Action::Up => write!(f, "Up"),
-            Action::PageDown => write!(f, "PageDown"),
-            Action::PageUp => write!(f, "PageUp"),
             Action::Enter => write!(f, "Enter"),
-            _ => write!(f, "{:?}", self),
+            _ => write!(f, "{self:?}"),
         }
     }
 }
