@@ -4,28 +4,33 @@
 // GNU General Public License version 2.
 
 use scx_utils::compat;
-use scxtop::available_kprobe_events;
 use scxtop::bpf_skel::types::bpf_event;
 use scxtop::cli::{generate_completions, Cli, Commands, TraceArgs, TuiArgs};
 use scxtop::config::Config;
 use scxtop::edm::{ActionHandler, BpfEventActionPublisher, BpfEventHandler, EventDispatchManager};
+use scxtop::layered_util;
 use scxtop::mangoapp::poll_mangoapp;
-use scxtop::read_file_string;
+use scxtop::search;
 use scxtop::tracer::Tracer;
+use scxtop::util::{get_clock_value, read_file_string};
 use scxtop::Action;
 use scxtop::App;
+use scxtop::CpuStatTracker;
 use scxtop::Event;
 use scxtop::Key;
 use scxtop::KeyMap;
+use scxtop::MemStatSnapshot;
 use scxtop::PerfettoTraceManager;
-use scxtop::Search;
+use scxtop::SystemStatAction;
 use scxtop::Tui;
 use scxtop::SCHED_NAME_PATH;
+use scxtop::{available_kprobe_events, UpdateColVisibilityAction};
 use scxtop::{bpf_skel::*, AppState};
 
 use anyhow::anyhow;
 use anyhow::Result;
 use clap::{CommandFactory, Parser};
+use futures::future::join_all;
 use libbpf_rs::skel::OpenSkel;
 use libbpf_rs::skel::SkelBuilder;
 use libbpf_rs::Link;
@@ -38,16 +43,16 @@ use ratatui::crossterm::event::{KeyCode::Char, KeyEvent};
 use simplelog::{
     ColorChoice, Config as SimplelogConfig, LevelFilter, TermLogger, TerminalMode, WriteLogger,
 };
-use std::sync::atomic::AtomicBool;
-use tokio::sync::mpsc;
-
 use std::ffi::CString;
 use std::fs::File;
 use std::mem::MaybeUninit;
 use std::str::FromStr;
+use std::sync::atomic::AtomicBool;
 use std::sync::atomic::Ordering;
 use std::sync::Arc;
 use std::time::Duration;
+use sysinfo::System;
+use tokio::sync::mpsc;
 
 fn get_action(app: &App, keymap: &KeyMap, event: Event) -> Action {
     match event {
@@ -57,24 +62,50 @@ fn get_action(app: &App, keymap: &KeyMap, event: Event) -> Action {
             Action::TickRateChange(std::time::Duration::from_millis(tick_rate_ms))
         }
         Event::Key(key) => handle_key_event(app, keymap, key),
-        Event::Paste(paste) => match app.state() {
-            AppState::PerfEvent | AppState::KprobeEvent => Action::InputEntry(paste),
-            _ => Action::None,
-        },
+        Event::Paste(paste) => handle_input_entry(app, paste).unwrap_or(Action::None),
         _ => Action::None,
     }
 }
 
 fn handle_key_event(app: &App, keymap: &KeyMap, key: KeyEvent) -> Action {
     match key.code {
-        Char(c) => match app.state() {
-            AppState::PerfEvent | AppState::KprobeEvent => Action::InputEntry(c.to_string()),
-            _ => keymap.action(&Key::Char(c)),
-        },
+        Char(c) => {
+            // Check if we should handle this character as input for filtering
+            if let Some(action) = handle_input_entry(app, c.to_string()) {
+                action
+            } else {
+                keymap.action(&Key::Char(c))
+            }
+        }
         _ => keymap.action(&Key::Code(key.code)),
     }
 }
 
+fn handle_input_entry(app: &App, s: String) -> Option<Action> {
+    match app.state() {
+        AppState::PerfEvent | AppState::KprobeEvent => Some(Action::InputEntry(s)),
+        AppState::Default
+        | AppState::Llc
+        | AppState::Node
+        | AppState::Process
+        | AppState::Memory => {
+            if app.filtering() {
+                Some(Action::InputEntry(s))
+            } else {
+                None
+            }
+        }
+        AppState::PerfTop => {
+            if app.filtering() {
+                Some(Action::InputEntry(s))
+            } else {
+                None
+            }
+        }
+        _ => None,
+    }
+}
+
 /// Attaches BPF programs to the skel.
 fn attach_progs(skel: &mut BpfSkel) -> Result<Vec<Link>> {
     // Attach probes
@@ -87,6 +118,9 @@ fn attach_progs(skel: &mut BpfSkel) -> Result<Vec<Link>> {
         skel.progs.on_sched_wakeup_new.attach()?,
         skel.progs.on_sched_waking.attach()?,
         skel.progs.on_sched_migrate_task.attach()?,
+        skel.progs.on_sched_fork.attach()?,
+        skel.progs.on_sched_exec.attach()?,
+        skel.progs.on_sched_exit.attach()?,
     ];
 
     // 6.13 compatibility
@@ -129,9 +163,6 @@ fn attach_progs(skel: &mut BpfSkel) -> Result<Vec<Link>> {
     if let Ok(link) = skel.progs.on_cpuhp_exit.attach() {
         links.push(link);
     }
-    if let Ok(link) = skel.progs.on_pstate_sample.attach() {
-        links.push(link);
-    }
 
     Ok(links)
 }
@@ -148,9 +179,9 @@ fn run_trace(trace_args: &TraceArgs) -> Result<()> {
         ColorChoice::Auto,
     )?;
 
-    let kprobe_events = Search::new(available_kprobe_events()?);
-    kprobe_events
-        .contains_all(&trace_args.kprobes)
+    let mut kprobe_events = available_kprobe_events()?;
+    kprobe_events.sort();
+    search::sorted_contains_all(&kprobe_events, &trace_args.kprobes)
         .then_some(())
         .ok_or_else(|| anyhow!("Invalid kprobe events"))?;
 
@@ -168,6 +199,7 @@ fn run_trace(trace_args: &TraceArgs) -> Result<()> {
         .block_on(async {
             let (action_tx, mut action_rx) = mpsc::unbounded_channel();
 
+            // Set up the BPF skel and publisher
             let mut open_object = MaybeUninit::uninit();
             let mut builder = BpfSkelBuilder::default();
             if trace_args.verbose > 2 {
@@ -175,22 +207,20 @@ fn run_trace(trace_args: &TraceArgs) -> Result<()> {
             }
 
             let skel = builder.open(&mut open_object)?;
-            // set sample rate to 1 here to populate the BPF tctxs
-            skel.maps.data_data.sample_rate = 1;
             compat::cond_kprobe_enable("gpu_memory_total", &skel.progs.on_gpu_memory_total)?;
             compat::cond_kprobe_enable("hw_pressure_update", &skel.progs.on_hw_pressure_update)?;
+            compat::cond_tracepoint_enable("sched:sched_process_wait", &skel.progs.on_sched_wait)?;
+            compat::cond_tracepoint_enable("sched:sched_process_hang", &skel.progs.on_sched_hang)?;
 
             let mut skel = skel.load()?;
-            let mut links = attach_progs(&mut skel)?;
-            links.push(skel.progs.on_sched_fork.attach()?);
-            links.push(skel.progs.on_sched_exec.attach()?);
-            links.push(skel.progs.on_sched_exit.attach()?);
+            skel.maps.data_data.as_mut().unwrap().enable_bpf_events = false;
+            let links = attach_progs(&mut skel)?;
 
             let bpf_publisher = BpfEventActionPublisher::new(action_tx.clone());
 
+            // Set up the event buffer
             let mut event_rbb = RingBufferBuilder::new();
             let mut edm = EventDispatchManager::new(None, None);
-            let warmup_done = Arc::new(AtomicBool::new(false));
             edm.register_bpf_handler(Box::new(bpf_publisher));
             let event_handler = move |data: &[u8]| {
                 let mut event = bpf_event::default();
@@ -198,67 +228,106 @@ fn run_trace(trace_args: &TraceArgs) -> Result<()> {
                 let _ = edm.on_event(&event);
                 0
             };
-
             event_rbb.add(&skel.maps.events, event_handler)?;
             let event_rb = event_rbb.build()?;
+
+            // Set up the background threads
             let shutdown = Arc::new(AtomicBool::new(false));
             let stop_poll = shutdown.clone();
-            let stop_main = shutdown.clone();
+            let stop_stats = shutdown.clone();
 
-            tokio::spawn(async move {
+            let mut handles = Vec::new();
+            handles.push(tokio::spawn(async move {
                 loop {
                     let _ = event_rb.poll(Duration::from_millis(1));
                     if stop_poll.load(Ordering::Relaxed) {
+                        // Flush the ring buffer to ensure all events are processed
+                        let _ = event_rb.consume();
                         debug!("polling stopped");
                         break;
                     }
                 }
-            });
+            }));
+
+            if trace_args.system_stats {
+                let mut cpu_stat_tracker = CpuStatTracker::default();
+                let mut mem_stats = MemStatSnapshot::default();
+                let mut system = System::new_all();
+                let action_tx_clone = action_tx.clone();
+
+                handles.push(tokio::spawn(async move {
+                    loop {
+                        if stop_stats.load(Ordering::Relaxed) {
+                            break;
+                        }
+                        let ts = get_clock_value(libc::CLOCK_BOOTTIME);
+
+                        cpu_stat_tracker
+                            .update(&mut system)
+                            .expect("Failed to update cpu stats");
+
+                        mem_stats.update().expect("Failed to update mem stats");
+
+                        let sys_stat_action = Action::SystemStat(SystemStatAction {
+                            ts,
+                            cpu_data_prev: cpu_stat_tracker.prev.clone(),
+                            cpu_data_current: cpu_stat_tracker.current.clone(),
+                            mem_info: mem_stats.clone(),
+                        });
+                        action_tx_clone
+                            .send(sys_stat_action)
+                            .expect("Failed to send CpuStat action");
+
+                        tokio::time::sleep(Duration::from_millis(100)).await;
+                    }
+                }));
+            }
 
-            let (complete_tx, mut complete_rx) = mpsc::channel(1);
             let trace_file_prefix = config.trace_file_prefix().to_string();
             let trace_file = trace_args.output_file.clone();
             let mut trace_manager = PerfettoTraceManager::new(trace_file_prefix, None);
+
+            info!("starting trace for {}ms", trace_args.trace_ms);
+            trace_manager.start()?;
             let mut tracer = Tracer::new(skel);
             tracer.trace(&trace_args.kprobes)?;
 
-            info!("warming up for {}ms", trace_args.warmup_ms);
-            tokio::time::sleep(Duration::from_millis(trace_args.warmup_ms)).await;
-            debug!("starting trace");
-            let thread_warmup = warmup_done.clone();
-            trace_manager.start()?;
-            thread_warmup.store(true, Ordering::Relaxed);
-            tokio::spawn(async move {
+            handles.push(tokio::spawn(async move {
                 let mut count = 0;
                 loop {
                     let action = action_rx.recv().await;
                     if let Some(a) = action {
-                        if thread_warmup.load(Ordering::Relaxed) {
-                            count += 1;
-                            trace_manager.on_action(&a).unwrap();
-                        }
-                    }
-                    if stop_main.load(Ordering::Relaxed) {
+                        count += 1;
+                        trace_manager
+                            .on_action(&a)
+                            .expect("Action should have been resolved");
+                    } else {
                         trace_manager.stop(trace_file, None).unwrap();
-                        info!("trace complete, collected {} events", count);
-                        let _ = complete_tx.send(1).await;
+                        info!("trace file compiled, collected {count} events");
                         break;
                     }
                 }
-            });
-
+            }));
             tokio::time::sleep(Duration::from_millis(trace_args.trace_ms)).await;
-            tracer.clear_links()?;
-            // The order is important here:
-            // 1) first drop the links to detach the attached BPF programs
-            // 2) set the shutdown variable to stop background tokio threads
-            // 3) wait for the completion of the trace file generation to complete
-            drop(links);
+
+            // 1) set the shutdown variable to stop background tokio threads
+            // 2) next, drop the links to detach the attached BPF programs
+            // 3) drop the action_tx to ensure action_rx closes
+            // 4) wait for the completion of the trace file generation to complete
             shutdown.store(true, Ordering::Relaxed);
-            let _ = complete_rx.recv().await;
+            tracer.clear_links()?;
+            drop(links);
+            drop(action_tx);
+            info!("generating trace");
+            let results = join_all(handles).await;
+            for result in results {
+                if let Err(e) = result {
+                    eprintln!("Task panicked: {e}");
+                }
+            }
 
             let stats = tracer.stats()?;
-            info!("{:?}", stats);
+            info!("{stats:?}");
 
             Ok(())
         })
@@ -305,19 +374,38 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
             let mut edm = EventDispatchManager::new(None, None);
             edm.register_bpf_handler(Box::new(bpf_publisher));
 
-            let skel = builder.open(&mut open_object)?;
-            skel.maps.rodata_data.long_tail_tracing_min_latency_ns =
+            let mut skel = builder.open(&mut open_object)?;
+            skel.maps.rodata_data.as_mut().unwrap().long_tail_tracing_min_latency_ns =
                 tui_args.experimental_long_tail_tracing_min_latency_ns;
 
+            let map_handle = if tui_args.layered {
+                skel.maps.rodata_data.as_mut().unwrap().layered = true;
+                action_tx.send(Action::UpdateColVisibility(UpdateColVisibilityAction {
+                    table: "Process".to_string(),
+                    col: "Layer ID".to_string(),
+                    visible: true,
+                }))?;
+                action_tx.send(Action::UpdateColVisibility(UpdateColVisibilityAction {
+                    table: "Thread".to_string(),
+                    col: "Layer ID".to_string(),
+                    visible: true,
+                }))?;
+                Some(layered_util::attach_to_existing_map("task_ctxs", &mut skel.maps.task_ctxs)?)
+            } else {
+                None
+            };
+
             compat::cond_kprobe_enable("gpu_memory_total", &skel.progs.on_gpu_memory_total)?;
             compat::cond_kprobe_enable("hw_pressure_update", &skel.progs.on_hw_pressure_update)?;
+            compat::cond_tracepoint_enable("sched:sched_process_wait", &skel.progs.on_sched_wait)?;
+            compat::cond_tracepoint_enable("sched:sched_process_hang", &skel.progs.on_sched_hang)?;
             let mut skel = skel.load()?;
             let mut links = attach_progs(&mut skel)?;
             skel.progs.scxtop_init.test_run(ProgramInput::default())?;
 
             if tui_args.experimental_long_tail_tracing {
-                skel.maps.data_data.trace_duration_ns = config.trace_duration_ns();
-                skel.maps.data_data.trace_warmup_ns = config.trace_warmup_ns();
+                skel.maps.data_data.as_mut().unwrap().trace_duration_ns = config.trace_duration_ns();
+                skel.maps.data_data.as_mut().unwrap().trace_warmup_ns = config.trace_warmup_ns();
 
                 let binary = tui_args
                     .experimental_long_tail_tracing_binary
@@ -335,7 +423,7 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
                         0,
                         UprobeOpts {
                             retprobe: true,
-                            func_name: symbol.clone(),
+                            func_name: Some(symbol.clone()),
                             ..Default::default()
                         },
                     )?,
@@ -345,14 +433,14 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
                         0,
                         UprobeOpts {
                             retprobe: false,
-                            func_name: symbol.clone(),
+                            func_name: Some(symbol.clone()),
                             ..Default::default()
                         },
                     )?,
                 ]);
             };
 
-            let mut tui = Tui::new(keymap.clone(), config.tick_rate_ms())?;
+            let mut tui = Tui::new(keymap.clone(), config.tick_rate_ms(), config.frame_rate_ms())?;
             let mut event_rbb = RingBufferBuilder::new();
             let event_handler = move |data: &[u8]| {
                 let mut event = bpf_event::default();
@@ -369,6 +457,7 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
                 scheduler,
                 100,
                 tui_args.process_id,
+                tui_args.layered,
                 action_tx.clone(),
                 skel,
             )?;
@@ -401,7 +490,6 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
                 });
             }
 
-
             loop {
                 tokio::select! {
                     ev = tui.next() => {
@@ -416,7 +504,9 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
                                 if app.should_quit.load(Ordering::Relaxed) {
                                     break;
                                 }
-                                tui.draw(|f| app.render(f).expect("Failed to render application"))?;
+                                if app.state() != AppState::Pause {
+                                    tui.draw(|f| app.render(f).expect("Failed to render application"))?;
+                                }
                             }
                             Event::Key(_) => {
                                 let action = get_action(&app, &keymap, ev);
@@ -433,6 +523,7 @@ fn run_tui(tui_args: &TuiArgs) -> Result<()> {
             }
             tui.exit()?;
             drop(links);
+            drop(map_handle);
 
             Ok(())
         })
@@ -450,7 +541,7 @@ fn main() -> Result<()> {
         }
         Commands::GenerateCompletions { shell, output } => {
             generate_completions(Cli::command(), *shell, output.clone())
-                .unwrap_or_else(|_| panic!("Failed to generate completions for {}", shell));
+                .unwrap_or_else(|_| panic!("Failed to generate completions for {shell}"));
         }
     }
     Ok(())
